#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
from collections.abc import Callable, Sequence
from typing import Any

from osc_lib.i18n import _


class _CommaListAction(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        if not isinstance(values, str):
            raise TypeError('expected str')
        setattr(namespace, self.dest, values.split(','))


def add_tag_filtering_option_to_parser(
    parser: argparse.ArgumentParser,
    resource_name: str,
    enhance_help: Callable[[str], str] = lambda _h: _h,
) -> None:
    """Add tag filtering options to a parser.

    :param parser: argparse.Argument parser object.
    :param resource_name: Description of the object being filtered.
    :param enhance_help: A callable accepting a single parameter, the
        (translated) help string, and returning a (translated) help string. May
        be used by a caller wishing to add qualifying text, such as "Applies to
        version XYZ only", to the help strings for all options produced by this
        method.
    """
    parser.add_argument(
        '--tags',
        metavar='<tag>[,<tag>,...]',
        action=_CommaListAction,
        help=enhance_help(
            _(
                'List %s which have all given tag(s) '
                '(Comma-separated list of tags)'
            )
            % resource_name
        ),
    )
    parser.add_argument(
        '--any-tags',
        metavar='<tag>[,<tag>,...]',
        action=_CommaListAction,
        help=enhance_help(
            _(
                'List %s which have any given tag(s) '
                '(Comma-separated list of tags)'
            )
            % resource_name
        ),
    )
    parser.add_argument(
        '--not-tags',
        metavar='<tag>[,<tag>,...]',
        action=_CommaListAction,
        help=enhance_help(
            _(
                'Exclude %s which have all given tag(s) '
                '(Comma-separated list of tags)'
            )
            % resource_name
        ),
    )
    parser.add_argument(
        '--not-any-tags',
        metavar='<tag>[,<tag>,...]',
        action=_CommaListAction,
        help=enhance_help(
            _(
                'Exclude %s which have any given tag(s) '
                '(Comma-separated list of tags)'
            )
            % resource_name
        ),
    )


def get_tag_filtering_args(
    parsed_args: argparse.Namespace,
    args: dict[str, Any],
) -> None:
    """Adds the tag arguments to an args list.

    Intended to be used to append the tags to an argument list that will be
    used for service client.

    :param parsed_args: Parsed argument object returned by argparse parse_args.
    :param args: The argument list to add tags to.
    """
    if parsed_args.tags:
        args['tags'] = ','.join(parsed_args.tags)
    if parsed_args.any_tags:
        args['any_tags'] = ','.join(parsed_args.any_tags)
    if parsed_args.not_tags:
        args['not_tags'] = ','.join(parsed_args.not_tags)
    if parsed_args.not_any_tags:
        args['not_any_tags'] = ','.join(parsed_args.not_any_tags)


def add_tag_option_to_parser_for_create(
    parser: argparse.ArgumentParser,
    resource_name: str,
    enhance_help: Callable[[str], str] = lambda _h: _h,
) -> None:
    """Add tag options to a parser for create commands.

    :param parser: argparse.Argument parser object.
    :param resource_name: Description of the object being filtered.
    :param enhance_help: A callable accepting a single parameter, the
        (translated) help string, and returning a (translated) help string. May
        be used by a caller wishing to add qualifying text, such as "Applies to
        version XYZ only", to the help strings for all options produced by this
        method.
    """
    tag_group = parser.add_mutually_exclusive_group()
    tag_group.add_argument(
        '--tag',
        action='append',
        dest='tags',
        metavar='<tag>',
        help=enhance_help(
            _("Tag to be added to the %s (repeat option to set multiple tags)")
            % resource_name
        ),
    )
    tag_group.add_argument(
        '--no-tag',
        action='store_true',
        help=enhance_help(_("No tags associated with the %s") % resource_name),
    )


def add_tag_option_to_parser_for_set(
    parser: argparse.ArgumentParser,
    resource_name: str,
    enhance_help: Callable[[str], str] = lambda _h: _h,
) -> None:
    """Add tag options to a parser for set commands.

    :param parser: argparse.Argument parser object.
    :param resource_name: Description of the object being filtered.
    :param enhance_help: A callable accepting a single parameter, the
        (translated) help string, and returning a (translated) help string. May
        be used by a caller wishing to add qualifying text, such as "Applies to
        version XYZ only", to the help strings for all options produced by this
        method.
    """
    parser.add_argument(
        '--tag',
        action='append',
        dest='tags',
        metavar='<tag>',
        help=enhance_help(
            _("Tag to be added to the %s (repeat option to set multiple tags)")
            % resource_name
        ),
    )
    parser.add_argument(
        '--no-tag',
        action='store_true',
        help=enhance_help(
            _(
                "Clear tags associated with the %s. Specify both "
                "--tag and --no-tag to overwrite current tags"
            )
            % resource_name
        ),
    )


def add_tag_option_to_parser_for_unset(
    parser: argparse.ArgumentParser,
    resource_name: str,
    enhance_help: Callable[[str], str] = lambda _h: _h,
) -> None:
    """Add tag options to a parser for set commands.

    :param parser: argparse.Argument parser object.
    :param resource_name: Description of the object being filtered.
    :param enhance_help: A callable accepting a single parameter, the
        (translated) help string, and returning a (translated) help string. May
        be used by a caller wishing to add qualifying text, such as "Applies to
        version XYZ only", to the help strings for all options produced by this
        method.
    """
    tag_group = parser.add_mutually_exclusive_group()
    tag_group.add_argument(
        '--tag',
        action='append',
        dest='tags',
        metavar='<tag>',
        help=enhance_help(
            _(
                "Tag to be removed from the %s "
                "(repeat option to remove multiple tags)"
            )
            % resource_name
        ),
    )
    tag_group.add_argument(
        '--all-tag',
        action='store_true',
        help=enhance_help(
            _("Clear all tags associated with the %s") % resource_name
        ),
    )


def update_tags_for_set(
    client: Any, obj: Any, parsed_args: argparse.Namespace
) -> None:
    """Set the tags on an object.

    :param client: The service client to use setting the tags.
    :param obj: The object (Resource) to set the tags on.
    :param parsed_args: Parsed argument object returned by argparse parse_args.
    """
    if parsed_args.no_tag:
        tags = set()
    else:
        tags = set(obj.tags or [])
    if parsed_args.tags:
        tags |= set(parsed_args.tags)
    if set(obj.tags or []) != tags:
        client.set_tags(obj, sorted(list(tags)))


def update_tags_for_unset(
    client: Any, obj: Any, parsed_args: argparse.Namespace
) -> None:
    """Unset the tags on an object.

    :param client: The service client to use unsetting the tags.
    :param obj: The object (Resource) to unset the tags on.
    :param parsed_args: Parsed argument object returned by argparse parse_args.
    """
    tags = set(obj.tags)
    if parsed_args.all_tag:
        tags = set()
    if parsed_args.tags:
        tags -= set(parsed_args.tags)
    if set(obj.tags) != tags:
        client.set_tags(obj, sorted(list(tags)))



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import warnings

from osc_lib.test.base import TestCase, TestCommand
from osc_lib.test.fakes import (
    AUTH_URL,
    INTERFACE,
    IDENTITY_API_VERSION as VERSION,
    PASSWORD,
    PROJECT_ID,
    PROJECT_NAME,
    REGION_NAME,
    TOKEN as AUTH_TOKEN,
    USERNAME,
    USER_ID,
    FakeApp,
    FakeClientManager,
    FakeLog,
    FakeModule,
    FakeOptions,
    FakeResource,
    FakeStdout,
)
from osc_lib.tests.test_clientmanager import (
    SERVICE_PROVIDER_ID,
    TEST_RESPONSE_DICT,
    TEST_RESPONSE_DICT_V3,
    TEST_VERSIONS,
)

__all__ = [
    'AUTH_TOKEN',
    'AUTH_URL',
    'FakeApp',
    'FakeClientManager',
    'FakeLog',
    'FakeModule',
    'FakeOptions',
    'FakeResource',
    'FakeStdout',
    'INTERFACE',
    'PASSWORD',
    'PROJECT_ID',
    'PROJECT_NAME',
    'REGION_NAME',
    'SERVICE_PROVIDER_ID',
    'TEST_RESPONSE_DICT',
    'TEST_RESPONSE_DICT_V3',
    'TEST_VERSIONS',
    'TestCase',
    'TestCommand',
    'to_unicode_dict',
    'USER_ID',
    'USERNAME',
    'VERSION',
]


warnings.warn(
    "The osc_lib.tests.fakes module is deprecated for removal and should not "
    "be used externally. Use the osc_lib.test module instead, which provides "
    "a public API.",
    DeprecationWarning,
)


# NOTE(stephenfin): This isn't moved since it's unused now. We can delete it in
# the future.
def to_unicode_dict(catalog_dict):
    """Converts dict to unicode dict"""
    if isinstance(catalog_dict, dict):
        return {
            to_unicode_dict(key): to_unicode_dict(value)
            for key, value in catalog_dict.items()
        }
    elif isinstance(catalog_dict, list):
        return [to_unicode_dict(element) for element in catalog_dict]
    elif isinstance(catalog_dict, str):
        return catalog_dict + ""
    else:
        return catalog_dict



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""API Utilities Library"""

from typing import Any, TypeVar


_T = TypeVar('_T', bound=list[Any])


def simple_filter(
    data: _T | None = None,
    attr: str | None = None,
    value: str | None = None,
    property_field: str | None = None,
) -> _T | None:
    """Filter a list of dicts

    :param list data:
        The list to be filtered.  The list is modified in-place and will
        be changed if any filtering occurs.
    :param string attr:
        The name of the attribute to filter.  If attr does not exist no
        match will succeed and no rows will be returned.  If attr is
        None no filtering will be performed and all rows will be returned.
    :param string value:
        The value to filter.  None is considered to be a 'no filter' value.
        '' matches against a Python empty string.
    :param string property_field:
        The name of the data field containing a property dict to filter.
        If property_field is None, attr is a field name. If property_field
        is not None, attr is a property key name inside the named property
        field.

    :returns:
        Returns the filtered list
    :rtype list:

    This simple filter (one attribute, one exact-match value) searches a
    list of dicts to select items.  It first searches the item dict for a
    matching ``attr`` then does an exact-match on the ``value``.  If
    ``property_field`` is given, it will look inside that field (if it
    exists and is a dict) for a matching ``value``.
    """

    # Take the do-nothing case shortcut
    if not data or not attr or value is None:
        return data

    # NOTE:(dtroyer): This filter modifies the provided list in-place using
    # list.remove() so we need to start at the end so the loop pointer does
    # not skip any items after a deletion.
    for d in reversed(data):
        if attr in d:
            # Searching data fields
            search_value = d[attr]
        elif (
            property_field
            and property_field in d
            and isinstance(d[property_field], dict)
        ):
            # Searching a properties field - do this separately because
            # we don't want to fail over to checking the fields if a
            # property name is given.
            if attr in d[property_field]:
                search_value = d[property_field][attr]
            else:
                search_value = None
        else:
            search_value = None

        # could do regex here someday...
        if not search_value or search_value != value:
            # remove from list
            try:
                data.remove(d)
            except ValueError:
                # it's already gone!
                pass

    return data



#   Copyright 2017 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Formattable column for specify content type"""

from typing import Any

from cliff import columns

from osc_lib import utils


class DictColumn(columns.FormattableColumn[dict[str, Any]]):
    """Format column for dict content"""

    def human_readable(self) -> str:
        return utils.format_dict(self._value)

    def machine_readable(self) -> dict[str, Any]:
        return dict(self._value or {})


class DictListColumn(columns.FormattableColumn[dict[str, list[Any]]]):
    """Format column for dict, key is string, value is list"""

    def human_readable(self) -> str:
        return utils.format_dict_of_list(self._value) or ''

    def machine_readable(self) -> dict[str, list[Any]]:
        return dict(self._value or {})


class ListColumn(columns.FormattableColumn[list[Any]]):
    """Format column for list content"""

    def human_readable(self) -> str:
        return utils.format_list(self._value) or ''

    def machine_readable(self) -> list[Any]:
        return [x for x in self._value or []]


class ListDictColumn(columns.FormattableColumn[list[dict[str, Any]]]):
    """Format column for list of dict content"""

    def human_readable(self) -> str:
        return utils.format_list_of_dicts(self._value) or ''

    def machine_readable(self) -> list[dict[str, Any]]:
        return [dict(x) for x in self._value or []]


class SizeColumn(columns.FormattableColumn[int | float]):
    """Format column for file size content"""

    def human_readable(self) -> str:
        return utils.format_size(self._value)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
from collections.abc import Sequence
import os
from typing import Any
from unittest import mock

from cliff import columns as cliff_columns
from cliff import command as cliff_command
import fixtures
import testtools

from osc_lib.test import fakes


class ParserException(Exception):
    pass


class TestCase(testtools.TestCase):
    # provide additional context for failures
    maxDiff = None

    def setUp(self) -> None:
        super().setUp()

        if (
            os.environ.get("OS_STDOUT_CAPTURE") == "True"
            or os.environ.get("OS_STDOUT_CAPTURE") == "1"
        ):
            stdout = self.useFixture(fixtures.StringStream("stdout")).stream
            self.useFixture(fixtures.MonkeyPatch("sys.stdout", stdout))

        if (
            os.environ.get("OS_STDERR_CAPTURE") == "True"
            or os.environ.get("OS_STDERR_CAPTURE") == "1"
        ):
            stderr = self.useFixture(fixtures.StringStream("stderr")).stream
            self.useFixture(fixtures.MonkeyPatch("sys.stderr", stderr))

    def assertNotCalled(self, m: mock.Mock, msg: str | None = None) -> None:
        """Assert a function was not called"""
        if m.called:
            if not msg:
                msg = f'method {m} should not have been called'
            self.fail(msg)


class TestCommand(TestCase):
    """Test cliff command classes"""

    def setUp(self) -> None:
        super().setUp()
        # Build up a fake app
        self.fake_stdout = fakes.FakeStdout()
        self.fake_log = fakes.FakeLog()
        self.app = fakes.FakeApp(self.fake_stdout, self.fake_log)
        self.app.client_manager = fakes.FakeClientManager()

    def check_parser(
        self,
        cmd: cliff_command.Command,
        args: list[str],
        verify_args: list[tuple[str, Any]],
    ) -> argparse.Namespace:
        cmd_parser = cmd.get_parser('check_parser')
        try:
            parsed_args = cmd_parser.parse_args(args)
        except SystemExit:
            raise ParserException("Argument parse failed")
        for av in verify_args:
            attr, value = av
            if attr:
                self.assertIn(attr, parsed_args)
                self.assertEqual(value, getattr(parsed_args, attr))
        return parsed_args

    def assertItemEqual(
        self, expected: Sequence[Any], actual: Sequence[Any]
    ) -> None:
        """Compare item considering formattable columns.

        This method compares an observed item to an expected item column by
        column. If a column is a formattable column, observed and expected
        columns are compared using human_readable() and machine_readable().
        """
        self.assertEqual(len(expected), len(actual))
        for col_expected, col_actual in zip(expected, actual):
            if isinstance(col_expected, cliff_columns.FormattableColumn):
                self.assertIsInstance(col_actual, col_expected.__class__)
                self.assertEqual(
                    col_expected.human_readable(), col_actual.human_readable()
                )
                self.assertEqual(
                    col_expected.machine_readable(),
                    col_actual.machine_readable(),
                )
            else:
                self.assertEqual(col_expected, col_actual)

    def assertListItemEqual(
        self, expected: Sequence[Any], actual: Sequence[Any]
    ) -> None:
        """Compare a list of items considering formattable columns.

        Each pair of observed and expected items are compared
        using assertItemEqual() method.
        """
        self.assertEqual(len(expected), len(actual))
        for item_expected, item_actual in zip(expected, actual):
            self.assertItemEqual(item_expected, item_actual)



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Manage access to the clients, including authenticating when needed."""

import copy
import logging
from typing import Any, Protocol, cast
import warnings

from keystoneauth1 import access as ksa_access
from keystoneauth1.identity import base as ksa_identity_base
from keystoneauth1 import session as ksa_session
from openstack.config import cloud_region
from openstack import connection
from oslo_utils import strutils

from osc_lib.api import auth
from osc_lib import exceptions

LOG = logging.getLogger(__name__)


class ClientCache:
    """Descriptor class for caching created client handles."""

    def __init__(self, factory: Any) -> None:
        warnings.warn(
            "The ClientCache class is deprecated for removal as it has no "
            "users.",
            category=DeprecationWarning,
        )

        self.factory = factory
        self._handle = None

    def __get__(self, instance: Any, owner: Any) -> Any:
        # Tell the ClientManager to login to keystone
        if self._handle is None:
            try:
                self._handle = self.factory(instance)
            except AttributeError as err:
                # Make sure the failure propagates. Otherwise, the plugin just
                # quietly isn't there.
                raise exceptions.PluginAttributeError(err) from err
        return self._handle


class _PasswordHelper(Protocol):
    def __call__(self, prompt: str | None = None) -> str: ...


class ClientManager:
    """Manages access to API clients, including authentication."""

    session: ksa_session.Session

    # NOTE(dtroyer): Keep around the auth required state of the _current_
    #                command since ClientManager has no visibility to the
    #                command itself; assume auth is not required.
    _auth_required = False

    def __init__(
        self,
        cli_options: cloud_region.CloudRegion,
        api_version: dict[str, str],
        pw_func: _PasswordHelper | None = None,
        app_name: str | None = None,
        app_version: str | None = None,
    ) -> None:
        """Set up a ClientManager

        :param cli_options:
            Options collected from the command-line, environment, or wherever
        :param api_version:
            Dict of API versions: key is API name, value is the version
        :param pw_func:
            Callback function for asking the user for a password.  The function
            takes an optional string for the prompt ('Password: ' on None) and
            returns a string containing the password
        :param app_name:
            The name of the application for passing through to the useragent
        :param app_version:
            The version of the application for passing through to the useragent
        """

        self._cli_options = cli_options
        self._api_version = api_version
        self._pw_callback = pw_func
        self._app_name = app_name
        self._app_version = app_version
        self.region_name = self._cli_options.region_name
        self.interface = self._cli_options.interface

        self.timing = self._cli_options.timing

        self._auth_ref: ksa_access.AccessInfo | None = None

        # self.verify is the Requests-compatible form
        # self.cacert is the form used by the legacy client libs
        # self.insecure is not needed, use 'not self.verify'
        self.cacert = None
        self.verify, self.cert = self._cli_options.get_requests_verify_args()
        # If there is a cacert and we're verifying, it'll be in the verify
        # argument.
        if not isinstance(self.verify, bool):
            self.cacert = self.verify

        # TODO(mordred) We also don't have any support for setting or passing
        # in api_timeout, which is set in occ defaults but we skip occ defaults
        # so set it here by hand and later we should potentially expose this
        # directly to osc
        self._cli_options.config['api_timeout'] = None

        # Get logging from root logger
        root_logger = logging.getLogger('')
        LOG.setLevel(root_logger.getEffectiveLevel())

        # NOTE(gyee): use this flag to indicate whether auth setup has already
        # been completed. If so, do not perform auth setup again. The reason
        # we need this flag is that we want to be able to perform auth setup
        # outside of auth_ref as auth_ref itself is a property. We can not
        # retrofit auth_ref to optionally skip scope check. Some operations
        # do not require a scoped token. In those cases, we call setup_auth
        # prior to dereferrencing auth_ref.
        self._auth_setup_completed = False

    def setup_auth(self) -> None:
        """Set up authentication

        This is deferred until authentication is actually attempted because
        it gets in the way of things that do not require auth.
        """
        if self._auth_setup_completed:
            return

        # Stash the selected auth type
        self.auth_plugin_name = self._cli_options.config['auth_type']

        # Basic option checking to avoid unhelpful error messages
        auth.check_valid_authentication_options(
            self._cli_options, self.auth_plugin_name
        )

        # Horrible hack alert...must handle prompt for null password if
        # password auth is requested.
        if (
            self.auth_plugin_name.endswith('password')
            and not self._cli_options.auth.get('password')
            and self._pw_callback is not None
        ):
            self._cli_options.auth['password'] = self._pw_callback()

        LOG.debug('Using auth plugin: %s', self.auth_plugin_name)
        LOG.debug(
            'Using parameters %s',
            strutils.mask_password(self._cli_options.auth),
        )
        self.auth = self._cli_options.get_auth()

        if self.auth is None:
            raise exceptions.AuthorizationFailure(
                'Failed to load authentication information'
            )

        if self._cli_options.service_provider:
            if not isinstance(self.auth, ksa_identity_base.BaseIdentityPlugin):
                raise exceptions.AuthorizationFailure(
                    f'keystone-to-keystone (k2k) federation requires use '
                    f'of an Identity-based authentication plugin. '
                    f'Got: {self.auth_plugin_name}'
                )

            self.auth = auth.get_keystone2keystone_auth(
                self.auth,
                self._cli_options.service_provider,
                self._cli_options.remote_project_id,
                self._cli_options.remote_project_name,
                self._cli_options.remote_project_domain_id,
                self._cli_options.remote_project_domain_name,
            )

        self.session = self._cli_options.get_session()

        self.sdk_connection = connection.Connection(config=self._cli_options)

        # We might get auth_ref from SDK cacnihg
        if hasattr(self.auth, 'auth_ref') and self.auth.auth_ref:
            self._auth_ref = self.auth.auth_ref

        self._auth_setup_completed = True

    def validate_scope(self) -> None:
        if not self._auth_ref:
            raise Exception('no authentication information')

        if self._auth_ref.project_id is not None:
            # We already have a project scope.
            return
        if self._auth_ref.domain_id is not None:
            # We already have a domain scope.
            return

        # We do not have a scoped token (and the user's default project scope
        # was not implied), so the client needs to be explicitly configured
        # with a scope.
        auth.check_valid_authorization_options(
            self._cli_options,
            self.auth_plugin_name,
        )

    @property
    def auth_ref(self) -> ksa_access.AccessInfo | None:
        """Dereference will trigger an auth if it hasn't already"""
        if (
            not self._auth_required
            or self._cli_options.config['auth_type'] == 'none'
        ):
            # Forcibly skip auth if we know we do not need it
            return None
        if not self._auth_ref:
            self.setup_auth()
            LOG.debug("Get auth_ref")
            # narrow type: we know this is set now due to the setup_auth call
            assert self.auth is not None
            self._auth_ref = self.auth.get_auth_ref(self.session)
        return self._auth_ref

    def _override_for(self, service_type: str) -> str | None:
        key = '{}_endpoint_override'.format(service_type.replace('-', '_'))
        return cast(str | None, self._cli_options.config.get(key))

    def is_service_available(self, service_type: str) -> bool | None:
        """Check if a service type is in the current Service Catalog"""
        # If there is an override, assume the service is available
        if self._override_for(service_type):
            return True
        # Trigger authentication necessary to discover endpoint
        if self.auth_ref:
            service_catalog = self.auth_ref.service_catalog
        else:
            service_catalog = None
        # Assume that the network endpoint is enabled.
        service_available = None
        if service_catalog:
            if service_type in service_catalog.get_endpoints():
                service_available = True
                LOG.debug("%s endpoint in service catalog", service_type)
            else:
                service_available = False
                LOG.debug("No %s endpoint in service catalog", service_type)
        else:
            LOG.debug("No service catalog")
        return service_available

    def get_endpoint_for_service_type(
        self,
        service_type: str,
        region_name: str | None = None,
        interface: str = 'public',
    ) -> str | None:
        """Return the endpoint URL for the service type."""
        # Overrides take priority unconditionally
        override = self._override_for(service_type)
        if override:
            return override

        if not interface:
            interface = 'public'
        # See if we are using password flow auth, i.e. we have a
        # service catalog to select endpoints from
        if self.auth_ref:
            endpoint = self.auth_ref.service_catalog.url_for(
                service_type=service_type,
                region_name=region_name,
                interface=interface,
            )
        elif self.auth:
            # Get the passed endpoint directly from the auth plugin
            endpoint = self.auth.get_endpoint(
                self.session,
                interface=interface,
            )
        return endpoint

    def get_configuration(self) -> dict[str, Any]:
        return copy.deepcopy(self._cli_options.config)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""OpenStackConfig subclass for argument compatibility"""

import logging
from typing import Any

from keystoneauth1 import loading as ksa_loading
from openstack.config import exceptions as sdk_exceptions
from openstack.config import loader as config
from oslo_utils import strutils


LOG = logging.getLogger(__name__)


# Subclass OpenStackConfig in order to munge config values
# before auth plugins are loaded
class OSC_Config(config.OpenStackConfig):
    def _auth_select_default_plugin(
        self, config: dict[str, Any]
    ) -> dict[str, Any]:
        """Select a default plugin based on supplied arguments

        Migrated from auth.select_auth_plugin()
        """

        identity_version = str(config.get('identity_api_version', ''))

        if config.get('username', None) and not config.get('auth_type', None):
            if identity_version == '3':
                config['auth_type'] = 'v3password'
            elif identity_version.startswith('2'):
                config['auth_type'] = 'v2password'
            else:
                # let keystoneauth figure it out itself
                config['auth_type'] = 'password'
        elif config.get('token', None) and not config.get('auth_type', None):
            if identity_version == '3':
                config['auth_type'] = 'v3token'
            elif identity_version.startswith('2'):
                config['auth_type'] = 'v2token'
            else:
                # let keystoneauth figure it out itself
                config['auth_type'] = 'token'
        else:
            # The ultimate default is similar to the original behaviour,
            # but this time with version discovery
            if not config.get('auth_type', None):
                config['auth_type'] = 'password'

        LOG.debug("Auth plugin {} selected".format(config['auth_type']))
        return config

    def _auth_v2_arguments(self, config: dict[str, Any]) -> dict[str, Any]:
        """Set up v2-required arguments from v3 info

        Migrated from auth.build_auth_params()
        """

        if 'auth_type' in config and config['auth_type'].startswith("v2"):
            if 'project_id' in config['auth']:
                config['auth']['tenant_id'] = config['auth']['project_id']
            if 'project_name' in config['auth']:
                config['auth']['tenant_name'] = config['auth']['project_name']
        return config

    def _auth_v2_ignore_v3(self, config: dict[str, Any]) -> dict[str, Any]:
        """Remove v3 arguments if present for v2 plugin

        Migrated from clientmanager.setup_auth()
        """
        # NOTE(hieulq): If USER_DOMAIN_NAME, USER_DOMAIN_ID, PROJECT_DOMAIN_ID
        # or PROJECT_DOMAIN_NAME is present and API_VERSION is 2.0, then
        # ignore all domain related configs.
        if not str(config.get('identity_api_version', '')).startswith('2'):
            return config

        if not config.get('auth_type') or not config['auth_type'].endswith(
            'password'
        ):
            return config

        domain_props = [
            'project_domain_id',
            'project_domain_name',
            'user_domain_id',
            'user_domain_name',
        ]
        for prop in domain_props:
            if config['auth'].pop(prop, None) is not None:
                if config.get('cloud'):
                    LOG.warning(
                        "Ignoring domain related config {} for {}"
                        "because identity API version is 2.0".format(
                            prop, config['cloud']
                        )
                    )
                else:
                    LOG.warning(
                        f"Ignoring domain related config {prop} because"
                        " identity API version is 2.0"
                    )
        return config

    def _auth_default_domain(self, config: dict[str, Any]) -> dict[str, Any]:
        """Set a default domain from available arguments

        Migrated from clientmanager.setup_auth()
        """

        identity_version = str(config.get('identity_api_version', ''))
        auth_type = config.get('auth_type', None)

        # TODO(mordred): This is a usability improvement that's broadly useful
        # We should port it back up into os-client-config.
        default_domain = config.get('default_domain', None)
        if (
            identity_version == '3'
            and (auth_type and not auth_type.startswith('v2'))
            and default_domain
        ):
            # NOTE(stevemar): If PROJECT_DOMAIN_ID or PROJECT_DOMAIN_NAME is
            # present, then do not change the behaviour. Otherwise, set the
            # PROJECT_DOMAIN_ID to 'OS_DEFAULT_DOMAIN' for better usability.
            if (
                auth_type in ("password", "v3password", "v3totp")
                and not config['auth'].get('project_domain_id')
                and not config['auth'].get('project_domain_name')
            ):
                config['auth']['project_domain_id'] = default_domain

            # NOTE(stevemar): If USER_DOMAIN_ID or USER_DOMAIN_NAME is present,
            # then do not change the behaviour. Otherwise, set the
            # USER_DOMAIN_ID to 'OS_DEFAULT_DOMAIN' for better usability.
            # NOTE(aloga): this should only be set if there is a username.
            # TODO(dtroyer): Move this to os-client-config after the plugin has
            # been loaded so we can check directly if the options are accepted.
            if (
                auth_type in ("password", "v3password", "v3totp")
                and not config['auth'].get('user_domain_id')
                and not config['auth'].get('user_domain_name')
            ):
                config['auth']['user_domain_id'] = default_domain
        return config

    def auth_config_hook(self, config: dict[str, Any]) -> dict[str, Any]:
        """Allow examination of config values before loading auth plugin

        OpenStackClient will override this to perform additional checks
        on auth_type.
        """

        config = self._auth_select_default_plugin(config)
        config = self._auth_v2_arguments(config)
        config = self._auth_v2_ignore_v3(config)
        config = self._auth_default_domain(config)

        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug(
                "auth_config_hook(): %s", strutils.mask_password(str(config))
            )
        return config

    def _validate_auth(
        self,
        config: dict[str, Any],
        loader: ksa_loading.BaseLoader[Any],
    ) -> dict[str, Any]:
        """Validate auth plugin arguments"""
        # May throw a keystoneauth1.exceptions.NoMatchingPlugin

        plugin_options = loader.get_options()

        msgs = []
        prompt_options = []
        for p_opt in plugin_options:
            # if it's in config, win, move it and kill it from config dict
            # if it's in config.auth but not in config we're good
            # deprecated loses to current
            # provided beats default, deprecated or not
            winning_value = self._find_winning_auth_value(p_opt, config)
            if not winning_value:
                winning_value = self._find_winning_auth_value(
                    p_opt, config['auth']
                )

            # if the plugin tells us that this value is required
            # then error if it's doesn't exist now
            if not winning_value and p_opt.required:
                msgs.append(
                    'Missing value {auth_key}'
                    ' required for auth plugin {plugin}'.format(
                        auth_key=p_opt.name,
                        plugin=config.get('auth_type'),
                    )
                )

            # Clean up after ourselves
            for opt in [p_opt.name] + [o.name for o in p_opt.deprecated]:
                opt = opt.replace('-', '_')
                config.pop(opt, None)
                config['auth'].pop(opt, None)

            if winning_value:
                # Prefer the plugin configuration dest value if the value's key
                # is marked as depreciated.
                if p_opt.dest is None:
                    config['auth'][p_opt.name.replace('-', '_')] = (
                        winning_value
                    )
                else:
                    config['auth'][p_opt.dest] = winning_value

            # See if this needs a prompting
            if (
                'prompt' in vars(p_opt)
                and p_opt.prompt is not None
                and p_opt.dest not in config['auth']
                and self._pw_callback is not None
            ):
                # Defer these until we know all required opts are present
                prompt_options.append(p_opt)

        if msgs:
            raise sdk_exceptions.OpenStackConfigException('\n'.join(msgs))

        if prompt_options:
            if self._pw_callback is None:
                # we have no way to get the missing options from the user, so
                # error out
                raise sdk_exceptions.OpenStackConfigException(
                    'Missing authentication information: {prompt_options}'
                )

            for p_opt in prompt_options:
                config['auth'][p_opt.dest] = self._pw_callback(p_opt.prompt)

        return config

    # TODO(stephenfin): Add type once we have typing for SDK
    def load_auth_plugin(self, config: dict[str, Any]) -> Any:
        """Get auth plugin and validate args"""

        loader = self._get_auth_loader(config)
        config = self._validate_auth(config, loader)
        auth_plugin = loader.load_from_options(**config['auth'])
        return auth_plugin



#   Copyright 2012-2013 OpenStack Foundation
#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import warnings

from osc_lib.test.base import ParserException, TestCase, TestCommand
from osc_lib.tests.test_clientmanager import (
    BaseTestClientManager as TestClientManager,
)
from osc_lib.tests.test_shell import (
    fake_execute,
    make_shell,
    opt2attr,
    opt2env,
    TestShell,
    EnvFixture,
)

__all__ = [
    'fake_execute',
    'make_shell',
    'opt2attr',
    'opt2env',
    'EnvFixture',
    'ParserException',
    'TestCase',
    'TestClientManager',
    'TestCommand',
    'TestShell',
]

warnings.warn(
    "The osc_lib.tests.utils module is deprecated for removal and should not "
    "be used externally. Use the osc_lib.test module instead, which provides "
    "a public API.",
    DeprecationWarning,
)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
from unittest import mock

from openstack import exceptions
from openstack.identity.v3 import project
import testtools

from osc_lib.cli import identity as cli_identity
from osc_lib.test import base


class IdentityUtilsTestCase(base.TestCase):
    def test_add_project_owner_option_to_parser(self):
        parser = argparse.ArgumentParser()
        cli_identity.add_project_owner_option_to_parser(parser)
        parsed_args = parser.parse_args(
            ['--project', 'project1', '--project-domain', 'domain1']
        )
        self.assertEqual('project1', parsed_args.project)
        self.assertEqual('domain1', parsed_args.project_domain)

    def test_find_project(self):
        sdk_connection = mock.Mock()
        sdk_find_project = sdk_connection.identity.find_project
        sdk_find_project.return_value = mock.sentinel.project1

        ret = cli_identity.find_project(sdk_connection, 'project1')
        self.assertEqual(mock.sentinel.project1, ret)
        sdk_find_project.assert_called_once_with(
            'project1', ignore_missing=False, domain_id=None
        )

    def test_find_project_with_domain(self):
        domain1 = mock.Mock()
        domain1.id = 'id-domain1'

        sdk_connection = mock.Mock()
        sdk_find_domain = sdk_connection.identity.find_domain
        sdk_find_domain.return_value = domain1
        sdk_find_project = sdk_connection.identity.find_project
        sdk_find_project.return_value = mock.sentinel.project1

        ret = cli_identity.find_project(sdk_connection, 'project1', 'domain1')
        self.assertEqual(mock.sentinel.project1, ret)
        sdk_find_domain.assert_called_once_with(
            'domain1', ignore_missing=False
        )
        sdk_find_project.assert_called_once_with(
            'project1', ignore_missing=False, domain_id='id-domain1'
        )

    def test_find_project_with_forbidden_exception(self):
        sdk_connection = mock.Mock()
        sdk_find_project = sdk_connection.identity.find_project
        exc = exceptions.HttpException()
        exc.status_code = 403
        sdk_find_project.side_effect = exc

        ret = cli_identity.find_project(sdk_connection, 'project1')

        self.assertIsInstance(ret, project.Project)
        self.assertEqual('project1', ret.id)
        self.assertEqual('project1', ret.name)

    def test_find_project_with_generic_exception(self):
        sdk_connection = mock.Mock()
        sdk_find_project = sdk_connection.identity.find_project
        exc = exceptions.HttpException()
        # Some value other than 403.
        exc.status_code = 499
        sdk_find_project.side_effect = exc

        with testtools.ExpectedException(exceptions.HttpException):
            cli_identity.find_project(sdk_connection, 'project1')



#   Copyright 2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""argparse Custom Actions"""

import argparse
from collections.abc import Callable, Iterable
from typing import Any, TypeVar
from collections.abc import Sequence

from osc_lib.i18n import _

_T = TypeVar('_T')


class KeyValueAction(argparse.Action):
    """A custom action to parse arguments as key=value pairs

    Ensures that ``dest`` is a dict and values are strings.
    """

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        if not isinstance(values, str):
            raise TypeError('expected str')

        # Make sure we have an empty dict rather than None
        if getattr(namespace, self.dest, None) is None:
            setattr(namespace, self.dest, {})

        # Add value if an assignment else remove it
        if '=' in values:
            values_list = values.split('=', 1)
            # NOTE(qtang): Prevent null key setting in property
            if '' == values_list[0]:
                msg = _("Property key must be specified: %s")
                raise argparse.ArgumentError(self, msg % str(values))
            else:
                getattr(namespace, self.dest, {}).update(dict([values_list]))
        else:
            msg = _("Expected 'key=value' type, but got: %s")
            raise argparse.ArgumentError(self, msg % str(values))


class KeyValueAppendAction(argparse.Action):
    """A custom action to parse arguments as key=value pairs

    Ensures that ``dest`` is a dict and values are lists of strings.
    """

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        if not isinstance(values, str):
            raise TypeError('expected str')

        # Make sure we have an empty dict rather than None
        if getattr(namespace, self.dest, None) is None:
            setattr(namespace, self.dest, {})

        # Add value if an assignment else remove it
        if '=' in values:
            key, value = values.split('=', 1)
            # NOTE(qtang): Prevent null key setting in property
            if '' == key:
                msg = _("Property key must be specified: %s")
                raise argparse.ArgumentError(self, msg % str(values))

            dest = getattr(namespace, self.dest)
            if key in dest:
                dest[key].append(value)
            else:
                dest[key] = [value]
        else:
            msg = _("Expected 'key=value' type, but got: %s")
            raise argparse.ArgumentError(self, msg % str(values))


class MultiKeyValueAction(argparse.Action):
    """A custom action to parse arguments as key1=value1,key2=value2 pairs

    Ensure that ``dest`` is a list. The list will finally contain multiple
    dicts, with key=value pairs in them.

    NOTE: The arguments string should be a comma separated key-value pairs.
    And comma(',') and equal('=') may not be used in the key or value.
    """

    def __init__(
        self,
        option_strings: Sequence[str],
        dest: str,
        nargs: int | str | None = None,
        required_keys: Sequence[str] | None = None,
        optional_keys: Sequence[str] | None = None,
        const: _T | None = None,
        default: _T | str | None = None,
        type: Callable[[str], _T] | None = None,
        choices: Iterable[_T] | None = None,
        required: bool = False,
        help: str | None = None,
        metavar: str | tuple[str, ...] | None = None,
    ) -> None:
        """Initialize the action object, and parse customized options

        Required keys and optional keys can be specified when initializing
        the action to enable the key validation. If none of them specified,
        the key validation will be skipped.

        :param required_keys: a list of required keys
        :param optional_keys: a list of optional keys
        """
        if nargs:
            msg = _("Parameter 'nargs' is not allowed, but got %s")
            raise ValueError(msg % nargs)

        super().__init__(
            option_strings,
            dest,
            nargs=nargs,
            const=const,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

        # required_keys: A list of keys that is required. None by default.
        if required_keys and not isinstance(required_keys, list):
            msg = _("'required_keys' must be a list")
            raise TypeError(msg)

        self.required_keys = set(required_keys or [])

        # optional_keys: A list of keys that is optional. None by default.
        if optional_keys and not isinstance(optional_keys, list):
            msg = _("'optional_keys' must be a list")
            raise TypeError(msg)
        self.optional_keys = set(optional_keys or [])

    def validate_keys(self, keys: Sequence[str]) -> None:
        """Validate the provided keys.

        :param keys: A list of keys to validate.
        """
        # Check key validation
        valid_keys = self.required_keys | self.optional_keys
        if valid_keys:
            invalid_keys = [k for k in keys if k not in valid_keys]
            if invalid_keys:
                msg = _(
                    "Invalid keys %(invalid_keys)s specified.\n"
                    "Valid keys are: %(valid_keys)s"
                )
                raise argparse.ArgumentError(
                    self,
                    msg
                    % {
                        'invalid_keys': ', '.join(invalid_keys),
                        'valid_keys': ', '.join(valid_keys),
                    },
                )

        if self.required_keys:
            missing_keys = [k for k in self.required_keys if k not in keys]
            if missing_keys:
                msg = _(
                    "Missing required keys %(missing_keys)s.\n"
                    "Required keys are: %(required_keys)s"
                )
                raise argparse.ArgumentError(
                    self,
                    msg
                    % {
                        'missing_keys': ', '.join(missing_keys),
                        'required_keys': ', '.join(self.required_keys),
                    },
                )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        if not isinstance(values, str):
            raise TypeError('expected str')

        # Make sure we have an empty list rather than None
        if getattr(namespace, self.dest, None) is None:
            setattr(namespace, self.dest, [])

        params: dict[str, str] = {}
        for kv in values.split(','):
            # Add value if an assignment else raise ArgumentError
            if '=' in kv:
                kv_list = kv.split('=', 1)
                # NOTE(qtang): Prevent null key setting in property
                if '' == kv_list[0]:
                    msg = _("Each property key must be specified: %s")
                    raise argparse.ArgumentError(self, msg % str(kv))
                else:
                    params.update(dict([kv_list]))
            else:
                msg = _(
                    "Expected comma separated 'key=value' pairs, but got: %s"
                )
                raise argparse.ArgumentError(self, msg % str(kv))

        # Check key validation
        self.validate_keys(list(params))

        # Update the dest dict
        getattr(namespace, self.dest, []).append(params)


class MultiKeyValueCommaAction(MultiKeyValueAction):
    """Custom action to parse arguments from a set of key=value pair

    Ensures that ``dest`` is a dict.
    Parses dict by separating comma separated string into individual values
    Ex. key1=val1,val2,key2=val3 => {"key1": "val1,val2", "key2": "val3"}
    """

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        """Overwrite the __call__ function of MultiKeyValueAction

        This is done to handle scenarios where we may have comma seperated
        data as a single value.
        """
        if not isinstance(values, str):
            msg = _("Invalid key=value pair, non-string value provided: %s")
            raise argparse.ArgumentError(self, msg % str(values))

        # Make sure we have an empty list rather than None
        if getattr(namespace, self.dest, None) is None:
            setattr(namespace, self.dest, [])

        params: dict[str, str] = {}
        key = ''
        for kv in values.split(','):
            # Add value if an assignment else raise ArgumentError
            if '=' in kv:
                kv_list = kv.split('=', 1)
                # NOTE(qtang): Prevent null key setting in property
                if '' == kv_list[0]:
                    msg = _("A key must be specified before '=': %s")
                    raise argparse.ArgumentError(self, msg % str(kv))
                else:
                    params.update(dict([kv_list]))
                key = kv_list[0]
            else:
                # If the ',' split does not have key=value pair, then it
                # means the current value is a part of the previous
                # key=value pair, so append it.
                try:
                    params[key] = f"{params[key]},{kv}"
                except KeyError:
                    msg = _("A key=value pair is required: %s")
                    raise argparse.ArgumentError(self, msg % str(kv))

        # Check key validation
        self.validate_keys(list(params))

        # Update the dest dict
        getattr(namespace, self.dest, []).append(params)


class RangeAction(argparse.Action):
    """A custom action to parse a single value or a range of values

    Parses single integer values or a range of integer values delimited
    by a colon and returns a tuple of integers:
    '4' sets ``dest`` to (4, 4)
    '6:9' sets ``dest`` to (6, 9)
    """

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        if not isinstance(values, str):
            msg = _("Invalid range, non-string value provided")
            raise argparse.ArgumentError(self, msg)

        range = values.split(':')
        if len(range) == 0:
            # Nothing passed, return a zero default
            setattr(namespace, self.dest, (0, 0))
        elif len(range) == 1:
            # Only a single value is present
            setattr(namespace, self.dest, (int(range[0]), int(range[0])))
        elif len(range) == 2:
            # Range of two values
            if int(range[0]) <= int(range[1]):
                setattr(namespace, self.dest, (int(range[0]), int(range[1])))
            else:
                msg = _("Invalid range, %(min)s is not less than %(max)s")
                raise argparse.ArgumentError(
                    self,
                    msg
                    % {
                        'min': range[0],
                        'max': range[1],
                    },
                )
        else:
            # Too many values
            msg = _("Invalid range, too many values")
            raise argparse.ArgumentError(self, msg)


class NonNegativeAction(argparse.Action):
    """A custom action to check whether the value is non-negative or not

    Ensures the value is >= 0.
    """

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        if not isinstance(values, str | int | float):
            msg = _("%s expected a non-negative integer")
            raise argparse.ArgumentError(self, msg % str(option_string))

        if int(values) >= 0:
            setattr(namespace, self.dest, values)
        else:
            msg = _("%s expected a non-negative integer")
            raise argparse.ArgumentError(self, msg % str(option_string))



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import time
from unittest import mock
import uuid

from cliff import columns as cliff_columns

from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib.test import base
from osc_lib.test import fakes
from osc_lib import utils

PASSWORD = "Pa$$w0rd"
WASSPORD = "Wa$$p0rd"
DROWSSAP = "dr0w$$aP"


class FakeOddballResource(fakes.FakeResource):
    def get(self, attr):
        """get() is needed for utils.find_resource()"""
        if attr == 'id':
            return self.id
        elif attr == 'name':
            return self.name
        else:
            return None


class TestUtils(base.TestCase):
    def _get_test_items(self):
        item1 = {'a': 1, 'b': 2}
        item2 = {'a': 1, 'b': 3}
        item3 = {'a': 2, 'b': 2}
        item4 = {'a': 2, 'b': 1}
        return [item1, item2, item3, item4]

    def test_find_min_match_no_sort(self):
        items = self._get_test_items()
        sort_str = None
        flair = {}
        expect_items = items
        self.assertEqual(
            expect_items,
            list(utils.find_min_match(items, sort_str, **flair)),
        )

    def test_find_min_match_no_flair(self):
        items = self._get_test_items()
        sort_str = 'b'
        flair = {}
        expect_items = [items[3], items[0], items[2], items[1]]
        self.assertEqual(
            expect_items,
            utils.find_min_match(items, sort_str, **flair),
        )

    def test_find_min_match_a2(self):
        items = self._get_test_items()
        sort_str = 'b'
        flair = {'a': 2}
        expect_items = [items[3], items[2]]
        self.assertEqual(
            expect_items,
            utils.find_min_match(items, sort_str, **flair),
        )

    def test_find_min_match_b2(self):
        items = self._get_test_items()
        sort_str = 'b'
        flair = {'b': 2}
        expect_items = [items[0], items[2], items[1]]
        self.assertEqual(
            expect_items,
            utils.find_min_match(items, sort_str, **flair),
        )

    def test_find_min_match_b5(self):
        items = self._get_test_items()
        sort_str = 'b'
        flair = {'b': 5}
        expect_items = []
        self.assertEqual(
            expect_items,
            utils.find_min_match(items, sort_str, **flair),
        )

    def test_find_min_match_a2_b2(self):
        items = self._get_test_items()
        sort_str = 'b'
        flair = {'a': 2, 'b': 2}
        expect_items = [items[2]]
        self.assertEqual(
            expect_items,
            utils.find_min_match(items, sort_str, **flair),
        )

    def test_get_password_good(self):
        with mock.patch("getpass.getpass", return_value=PASSWORD):
            mock_stdin = mock.Mock()
            mock_stdin.isatty = mock.Mock()
            mock_stdin.isatty.return_value = True
            self.assertEqual(PASSWORD, utils.get_password(mock_stdin))

    def test_get_password_bad_once(self):
        answers = [PASSWORD, WASSPORD, DROWSSAP, DROWSSAP]
        with mock.patch("getpass.getpass", side_effect=answers):
            mock_stdin = mock.Mock()
            mock_stdin.isatty = mock.Mock()
            mock_stdin.isatty.return_value = True
            self.assertEqual(DROWSSAP, utils.get_password(mock_stdin))

    def test_get_password_no_tty(self):
        mock_stdin = mock.Mock()
        mock_stdin.isatty = mock.Mock()
        mock_stdin.isatty.return_value = False
        self.assertRaises(
            exceptions.CommandError, utils.get_password, mock_stdin
        )

    def test_get_password_cntrl_d(self):
        with mock.patch("getpass.getpass", side_effect=EOFError()):
            mock_stdin = mock.Mock()
            mock_stdin.isatty = mock.Mock()
            mock_stdin.isatty.return_value = True
            self.assertRaises(
                exceptions.CommandError, utils.get_password, mock_stdin
            )

    def test_sort_items_with_one_key(self):
        items = self._get_test_items()
        sort_str = 'b'
        expect_items = [items[3], items[0], items[2], items[1]]
        self.assertEqual(expect_items, utils.sort_items(items, sort_str))

    def test_sort_items_with_multiple_keys(self):
        items = self._get_test_items()
        sort_str = 'a,b'
        expect_items = [items[0], items[1], items[3], items[2]]
        self.assertEqual(expect_items, utils.sort_items(items, sort_str))

    def test_sort_items_all_with_direction(self):
        items = self._get_test_items()
        sort_str = 'a:desc,b:desc'
        expect_items = [items[2], items[3], items[1], items[0]]
        self.assertEqual(expect_items, utils.sort_items(items, sort_str))

    def test_sort_items_some_with_direction(self):
        items = self._get_test_items()
        sort_str = 'a,b:desc'
        expect_items = [items[1], items[0], items[2], items[3]]
        self.assertEqual(expect_items, utils.sort_items(items, sort_str))

    def test_sort_items_with_object(self):
        item1 = mock.Mock(a=1, b=2)
        item2 = mock.Mock(a=1, b=3)
        item3 = mock.Mock(a=2, b=2)
        item4 = mock.Mock(a=2, b=1)
        items = [item1, item2, item3, item4]
        sort_str = 'b,a'
        expect_items = [item4, item1, item3, item2]
        self.assertEqual(expect_items, utils.sort_items(items, sort_str))

    def test_sort_items_with_empty_key(self):
        items = self._get_test_items()
        sort_srt = ''
        self.assertEqual(items, utils.sort_items(items, sort_srt))
        sort_srt = None
        self.assertEqual(items, utils.sort_items(items, sort_srt))

    def test_sort_items_with_invalid_key(self):
        items = self._get_test_items()
        sort_str = 'c'
        self.assertRaises(
            exceptions.CommandError, utils.sort_items, items, sort_str
        )

    def test_sort_items_with_invalid_direction(self):
        items = self._get_test_items()
        sort_str = 'a:bad_dir'
        self.assertRaises(
            exceptions.CommandError, utils.sort_items, items, sort_str
        )

    def test_sort_items_with_different_type_exception(self):
        item1 = {'a': 2}
        item2 = {'a': 3}
        item3 = {'a': None}
        item4 = {'a': 1}
        items = [item1, item2, item3, item4]
        sort_str = 'a'
        self.assertRaises(TypeError, utils.sort_items, items, sort_str)

    def test_sort_items_with_different_type_int(self):
        item1 = {'a': 2}
        item2 = {'a': 3}
        item3 = {'a': None}
        item4 = {'a': 1}
        items = [item1, item2, item3, item4]
        sort_str = 'a'
        sort_type = int
        expect_items = [item3, item4, item1, item2]
        self.assertEqual(
            expect_items, utils.sort_items(items, sort_str, sort_type)
        )

    def test_sort_items_with_different_type_str(self):
        item1 = {'a': 'a'}
        item2 = {'a': None}
        item3 = {'a': '2'}
        item4 = {'a': 'b'}
        items = [item1, item2, item3, item4]
        sort_str = 'a'
        sort_type = str
        expect_items = [item3, item2, item1, item4]
        self.assertEqual(
            expect_items, utils.sort_items(items, sort_str, sort_type)
        )

    @mock.patch.object(time, 'sleep')
    def test_wait_for_delete_ok(self, mock_sleep):
        # Tests the normal flow that the resource is deleted with a 404 coming
        # back on the 2nd iteration of the wait loop.
        resource = mock.MagicMock(status='ACTIVE', progress=None)
        mock_get = mock.Mock(side_effect=[resource, exceptions.NotFound(404)])
        manager = mock.MagicMock(get=mock_get)
        res_id = str(uuid.uuid4())
        callback = mock.Mock()
        self.assertTrue(
            utils.wait_for_delete(manager, res_id, callback=callback)
        )
        mock_sleep.assert_called_once_with(5)
        callback.assert_called_once_with(0)

    @mock.patch.object(time, 'sleep')
    def test_wait_for_delete_timeout(self, mock_sleep):
        # Tests that we fail if the resource is not deleted before the timeout.
        resource = mock.MagicMock(status='ACTIVE')
        mock_get = mock.Mock(return_value=resource)
        manager = mock.MagicMock(get=mock_get)
        res_id = str(uuid.uuid4())
        self.assertFalse(
            utils.wait_for_delete(manager, res_id, sleep_time=1, timeout=1)
        )
        mock_sleep.assert_called_once_with(1)

    @mock.patch.object(time, 'sleep')
    def test_wait_for_delete_error(self, mock_sleep):
        # Tests that we fail if the resource goes to error state while waiting.
        resource = mock.MagicMock(status='ERROR')
        mock_get = mock.Mock(return_value=resource)
        manager = mock.MagicMock(get=mock_get)
        res_id = str(uuid.uuid4())
        self.assertFalse(utils.wait_for_delete(manager, res_id))
        mock_sleep.assert_not_called()

    @mock.patch.object(time, 'sleep')
    def test_wait_for_delete_error_with_overrides(self, mock_sleep):
        # Tests that we fail if the resource is my_status=failed
        resource = mock.MagicMock(my_status='FAILED')
        mock_get = mock.Mock(return_value=resource)
        manager = mock.MagicMock(get=mock_get)
        res_id = str(uuid.uuid4())
        self.assertFalse(
            utils.wait_for_delete(
                manager,
                res_id,
                status_field='my_status',
                error_status=['failed'],
            )
        )
        mock_sleep.assert_not_called()

    @mock.patch.object(time, 'sleep')
    def test_wait_for_delete_error_with_overrides_exception(self, mock_sleep):
        # Tests that we succeed if the resource is specific exception
        mock_get = mock.Mock(side_effect=Exception)
        manager = mock.MagicMock(get=mock_get)
        res_id = str(uuid.uuid4())
        self.assertTrue(
            utils.wait_for_delete(
                manager, res_id, exception_name=['Exception']
            )
        )
        mock_sleep.assert_not_called()

    @mock.patch.object(time, 'sleep')
    def test_wait_for_status_ok(self, mock_sleep):
        # Tests the normal flow that the resource is status=active
        resource = mock.MagicMock(status='ACTIVE')
        status_f = mock.Mock(return_value=resource)
        res_id = str(uuid.uuid4())
        self.assertTrue(
            utils.wait_for_status(
                status_f,
                res_id,
            )
        )
        mock_sleep.assert_not_called()

    @mock.patch.object(time, 'sleep')
    def test_wait_for_status_ok_with_overrides(self, mock_sleep):
        # Tests the normal flow that the resource is status=complete
        resource = mock.MagicMock(my_status='COMPLETE')
        status_f = mock.Mock(return_value=resource)
        res_id = str(uuid.uuid4())
        self.assertTrue(
            utils.wait_for_status(
                status_f,
                res_id,
                status_field='my_status',
                success_status=['complete'],
            )
        )
        mock_sleep.assert_not_called()

    @mock.patch.object(time, 'sleep')
    def test_wait_for_status_error(self, mock_sleep):
        # Tests that we fail if the resource is status=error
        resource = mock.MagicMock(status='ERROR')
        status_f = mock.Mock(return_value=resource)
        res_id = str(uuid.uuid4())
        self.assertFalse(utils.wait_for_status(status_f, res_id))
        mock_sleep.assert_not_called()

    @mock.patch.object(time, 'sleep')
    def test_wait_for_status_error_with_overrides(self, mock_sleep):
        # Tests that we fail if the resource is my_status=failed
        resource = mock.MagicMock(my_status='FAILED')
        status_f = mock.Mock(return_value=resource)
        res_id = str(uuid.uuid4())
        self.assertFalse(
            utils.wait_for_status(
                status_f,
                res_id,
                status_field='my_status',
                error_status=['failed'],
            )
        )
        mock_sleep.assert_not_called()

    def test_build_kwargs_dict_value_set(self):
        self.assertEqual(
            {'arg_bla': 'bla'}, utils.build_kwargs_dict('arg_bla', 'bla')
        )

    def test_build_kwargs_dict_value_None(self):
        self.assertEqual({}, utils.build_kwargs_dict('arg_bla', None))

    def test_build_kwargs_dict_value_empty_str(self):
        self.assertEqual({}, utils.build_kwargs_dict('arg_bla', ''))

    def test_is_ascii_bytes(self):
        self.assertFalse(utils.is_ascii(b'\xe2'))

    def test_is_ascii_string(self):
        self.assertFalse(utils.is_ascii('\u2665'))

    def test_format_size(self):
        self.assertEqual("999", utils.format_size(999))
        self.assertEqual("100K", utils.format_size(100000))
        self.assertEqual("2M", utils.format_size(2000000))
        self.assertEqual("16.4M", utils.format_size(16361280))
        self.assertEqual("1.6G", utils.format_size(1576395005))
        self.assertEqual("0", utils.format_size(None))

    def test_backward_compat_col_lister(self):
        fake_col_headers = ['ID', 'Name', 'Size']
        columns = ['Display Name']
        column_map = {'Display Name': 'Name'}
        results = utils.backward_compat_col_lister(
            fake_col_headers, columns, column_map
        )
        self.assertIsInstance(results, list)
        self.assertIn('Display Name', results)
        self.assertNotIn('Name', results)
        self.assertIn('ID', results)
        self.assertIn('Size', results)

    def test_backward_compat_col_lister_no_specify_column(self):
        fake_col_headers = ['ID', 'Name', 'Size']
        columns = []
        column_map = {'Display Name': 'Name'}
        results = utils.backward_compat_col_lister(
            fake_col_headers, columns, column_map
        )
        self.assertIsInstance(results, list)
        self.assertNotIn('Display Name', results)
        self.assertIn('Name', results)
        self.assertIn('ID', results)
        self.assertIn('Size', results)

    def test_backward_compat_col_lister_with_tuple_headers(self):
        fake_col_headers = ('ID', 'Name', 'Size')
        columns = ['Display Name']
        column_map = {'Display Name': 'Name'}
        results = utils.backward_compat_col_lister(
            fake_col_headers, columns, column_map
        )
        self.assertIsInstance(results, list)
        self.assertIn('Display Name', results)
        self.assertNotIn('Name', results)
        self.assertIn('ID', results)
        self.assertIn('Size', results)

    def test_backward_compat_col_showone(self):
        fake_object = {
            'id': 'fake-id',
            'name': 'fake-name',
            'size': 'fake-size',
        }
        columns = ['display_name']
        column_map = {'display_name': 'name'}
        results = utils.backward_compat_col_showone(
            fake_object, columns, column_map
        )
        self.assertIsInstance(results, dict)
        self.assertIn('display_name', results)
        self.assertIn('id', results)
        self.assertNotIn('name', results)
        self.assertIn('size', results)

    def test_backward_compat_col_showone_no_specify_column(self):
        fake_object = {
            'id': 'fake-id',
            'name': 'fake-name',
            'size': 'fake-size',
        }
        columns = []
        column_map = {'display_name': 'name'}
        results = utils.backward_compat_col_showone(
            fake_object, columns, column_map
        )
        self.assertIsInstance(results, dict)
        self.assertNotIn('display_name', results)
        self.assertIn('id', results)
        self.assertIn('name', results)
        self.assertIn('size', results)

    def _test_get_item_properties_with_formatter(self, formatters):
        names = ('id', 'attr')
        item = fakes.FakeResource(info={'id': 'fake-id', 'attr': ['a', 'b']})
        res_id, res_attr = utils.get_item_properties(
            item, names, formatters=formatters
        )
        self.assertEqual('fake-id', res_id)
        return res_attr

    def test_get_item_properties_with_formattable_column(self):
        formatters = {'attr': format_columns.ListColumn}
        res_attr = self._test_get_item_properties_with_formatter(formatters)
        self.assertIsInstance(res_attr, format_columns.ListColumn)

    def _test_get_dict_properties_with_formatter(self, formatters):
        names = ('id', 'attr')
        item = {'id': 'fake-id', 'attr': ['a', 'b']}
        res_id, res_attr = utils.get_dict_properties(
            item, names, formatters=formatters
        )
        self.assertEqual('fake-id', res_id)
        return res_attr

    def test_get_dict_properties_with_formattable_column(self):
        formatters = {'attr': format_columns.ListColumn}
        res_attr = self._test_get_dict_properties_with_formatter(formatters)
        self.assertIsInstance(res_attr, format_columns.ListColumn)

    def _test_calculate_header_and_attrs(
        self, parsed_args_columns, expected_headers, expected_attrs
    ):
        column_headers = ('ID', 'Name', 'Fixed IP Addresses')
        columns = ('id', 'name', 'fixed_ips')
        parsed_args = mock.Mock()
        parsed_args.columns = parsed_args_columns
        ret_headers, ret_attrs = utils.calculate_header_and_attrs(
            column_headers, columns, parsed_args
        )
        self.assertEqual(expected_headers, ret_headers)
        self.assertEqual(expected_attrs, ret_attrs)
        if parsed_args_columns:
            self.assertEqual(expected_headers, parsed_args.columns)
        else:
            self.assertFalse(parsed_args.columns)

    def test_calculate_header_and_attrs_without_column_arg(self):
        self._test_calculate_header_and_attrs(
            [],
            ('ID', 'Name', 'Fixed IP Addresses'),
            ('id', 'name', 'fixed_ips'),
        )

    def test_calculate_header_and_attrs_with_known_columns(self):
        self._test_calculate_header_and_attrs(
            ['Name', 'ID'], ['Name', 'ID'], ['name', 'id']
        )

    def test_calculate_header_and_attrs_with_unknown_columns(self):
        self._test_calculate_header_and_attrs(
            ['Name', 'ID', 'device_id'],
            ['Name', 'ID', 'device_id'],
            ['name', 'id', 'device_id'],
        )

    def test_calculate_header_and_attrs_with_attrname_columns(self):
        self._test_calculate_header_and_attrs(
            ['name', 'id', 'device_id'],
            ['Name', 'ID', 'device_id'],
            ['name', 'id', 'device_id'],
        )

    def test_subtest(self):
        for i in range(3):
            with self.subTest(i=i):
                self.assertEqual(i, i)


class NoUniqueMatch(Exception):
    pass


class TestFindResource(base.TestCase):
    def setUp(self):
        super().setUp()
        self.name = 'legos'
        self.expected = mock.Mock()
        self.manager = mock.Mock()
        self.manager.resource_class = mock.Mock()
        self.manager.resource_class.__name__ = 'lego'

    def test_find_resource_get_int(self):
        self.manager.get = mock.Mock(return_value=self.expected)
        result = utils.find_resource(self.manager, 1)
        self.assertEqual(self.expected, result)
        self.manager.get.assert_called_with(1)

    def test_find_resource_get_int_string(self):
        self.manager.get = mock.Mock(return_value=self.expected)
        result = utils.find_resource(self.manager, "2")
        self.assertEqual(self.expected, result)
        self.manager.get.assert_called_with("2")

    def test_find_resource_get_name_and_domain(self):
        name = 'admin'
        domain_id = '30524568d64447fbb3fa8b7891c10dd6'
        # NOTE(stevemar): we need an iterable side-effect because the same
        # function (manager.get()) is used twice, the first time an exception
        # will happen, then the result will be found, but only after using
        # the domain ID as a query arg
        side_effect = [Exception('Boom!'), self.expected]
        self.manager.get = mock.Mock(side_effect=side_effect)
        result = utils.find_resource(self.manager, name, domain_id=domain_id)
        self.assertEqual(self.expected, result)
        self.manager.get.assert_called_with(name, domain_id=domain_id)

    def test_find_resource_get_uuid(self):
        uuid = '9a0dc2a0-ad0d-11e3-a5e2-0800200c9a66'
        self.manager.get = mock.Mock(return_value=self.expected)
        result = utils.find_resource(self.manager, uuid)
        self.assertEqual(self.expected, result)
        self.manager.get.assert_called_with(uuid)

    def test_find_resource_get_whatever(self):
        self.manager.get = mock.Mock(return_value=self.expected)
        result = utils.find_resource(self.manager, 'whatever')
        self.assertEqual(self.expected, result)
        self.manager.get.assert_called_with('whatever')

    def test_find_resource_find(self):
        self.manager.get = mock.Mock(side_effect=Exception('Boom!'))
        self.manager.find = mock.Mock(return_value=self.expected)
        result = utils.find_resource(self.manager, self.name)
        self.assertEqual(self.expected, result)
        self.manager.get.assert_called_with(self.name)
        self.manager.find.assert_called_with(name=self.name)

    def test_find_resource_find_not_found(self):
        self.manager.get = mock.Mock(side_effect=Exception('Boom!'))
        self.manager.find = mock.Mock(
            side_effect=exceptions.NotFound(404, "2")
        )
        result = self.assertRaises(
            exceptions.CommandError,
            utils.find_resource,
            self.manager,
            self.name,
        )
        self.assertEqual(
            "No lego with a name or ID of 'legos' exists.", str(result)
        )
        self.manager.get.assert_called_with(self.name)
        self.manager.find.assert_called_with(name=self.name)

    def test_find_resource_list_forbidden(self):
        self.manager.get = mock.Mock(side_effect=Exception('Boom!'))
        self.manager.find = mock.Mock(side_effect=Exception('Boom!'))
        self.manager.list = mock.Mock(side_effect=exceptions.Forbidden(403))
        self.assertRaises(
            exceptions.Forbidden, utils.find_resource, self.manager, self.name
        )
        self.manager.list.assert_called_with()

    def test_find_resource_find_no_unique(self):
        self.manager.get = mock.Mock(side_effect=Exception('Boom!'))
        self.manager.find = mock.Mock(side_effect=NoUniqueMatch())
        result = self.assertRaises(
            exceptions.CommandError,
            utils.find_resource,
            self.manager,
            self.name,
        )
        self.assertEqual(
            "More than one lego exists with the name 'legos'.", str(result)
        )
        self.manager.get.assert_called_with(self.name)
        self.manager.find.assert_called_with(name=self.name)

    def test_find_resource_silly_resource(self):
        # We need a resource with no resource_class for this test, start fresh
        self.manager = mock.Mock()
        self.manager.get = mock.Mock(side_effect=Exception('Boom!'))
        self.manager.find = mock.Mock(
            side_effect=AttributeError(
                "'Controller' object has no attribute 'find'",
            )
        )
        silly_resource = FakeOddballResource(
            None,
            {'id': '12345', 'name': self.name},
            loaded=True,
        )
        self.manager.list = mock.Mock(
            return_value=[
                silly_resource,
            ],
        )
        result = utils.find_resource(self.manager, self.name)
        self.assertEqual(silly_resource, result)
        self.manager.get.assert_called_with(self.name)
        self.manager.find.assert_called_with(name=self.name)

    def test_find_resource_silly_resource_not_found(self):
        # We need a resource with no resource_class for this test, start fresh
        self.manager = mock.Mock()
        self.manager.get = mock.Mock(side_effect=Exception('Boom!'))
        self.manager.find = mock.Mock(
            side_effect=AttributeError(
                "'Controller' object has no attribute 'find'",
            )
        )
        self.manager.list = mock.Mock(return_value=[])
        result = self.assertRaises(
            exceptions.CommandError,
            utils.find_resource,
            self.manager,
            self.name,
        )
        self.assertEqual("Could not find resource legos", str(result))
        self.manager.get.assert_called_with(self.name)
        self.manager.find.assert_called_with(name=self.name)

    def test_find_resource_silly_resource_no_unique_match(self):
        # We need a resource with no resource_class for this test, start fresh
        self.manager = mock.Mock()
        self.manager.get = mock.Mock(side_effect=Exception('Boom!'))
        self.manager.find = mock.Mock(
            side_effect=AttributeError(
                "'Controller' object has no attribute 'find'",
            )
        )
        silly_resource = FakeOddballResource(
            None,
            {'id': '12345', 'name': self.name},
            loaded=True,
        )
        silly_resource_same = FakeOddballResource(
            None,
            {'id': 'abcde', 'name': self.name},
            loaded=True,
        )
        self.manager.list = mock.Mock(
            return_value=[silly_resource, silly_resource_same]
        )
        result = self.assertRaises(
            exceptions.CommandError,
            utils.find_resource,
            self.manager,
            self.name,
        )
        self.assertEqual(
            "More than one resource exists with the name or ID 'legos'.",
            str(result),
        )
        self.manager.get.assert_called_with(self.name)
        self.manager.find.assert_called_with(name=self.name)

    def test_format_dict(self):
        expected = "a='b', c='d', e='f'"
        self.assertEqual(
            expected, utils.format_dict({'a': 'b', 'c': 'd', 'e': 'f'})
        )
        self.assertEqual(
            expected, utils.format_dict({'e': 'f', 'c': 'd', 'a': 'b'})
        )
        self.assertIsNone(utils.format_dict(None))

    def test_format_dict_recursive(self):
        expected = "a='b', c.1='d', c.2=''"
        self.assertEqual(
            expected, utils.format_dict({'a': 'b', 'c': {'1': 'd', '2': ''}})
        )
        self.assertEqual(
            expected, utils.format_dict({'c': {'1': 'd', '2': ''}, 'a': 'b'})
        )
        self.assertIsNone(utils.format_dict(None))

        expected = "a1='A', a2.b1.c1='B', a2.b1.c2=, a2.b2='D'"
        self.assertEqual(
            expected,
            utils.format_dict(
                {
                    'a1': 'A',
                    'a2': {
                        'b1': {
                            'c1': 'B',
                            'c2': None,
                        },
                        'b2': 'D',
                    },
                }
            ),
        )
        self.assertEqual(
            expected,
            utils.format_dict(
                {
                    'a2': {
                        'b1': {
                            'c2': None,
                            'c1': 'B',
                        },
                        'b2': 'D',
                    },
                    'a1': 'A',
                }
            ),
        )

    def test_format_dict_of_list(self):
        expected = "a=a1, a2; b=b1, b2; c=c1, c2; e="
        self.assertEqual(
            expected,
            utils.format_dict_of_list(
                {
                    'a': ['a2', 'a1'],
                    'b': ['b2', 'b1'],
                    'c': ['c1', 'c2'],
                    'd': None,
                    'e': [],
                }
            ),
        )
        self.assertEqual(
            expected,
            utils.format_dict_of_list(
                {
                    'c': ['c1', 'c2'],
                    'a': ['a2', 'a1'],
                    'b': ['b2', 'b1'],
                    'e': [],
                }
            ),
        )
        self.assertIsNone(utils.format_dict_of_list(None))

    def test_format_dict_of_list_with_separator(self):
        expected = "a=a1, a2\nb=b1, b2\nc=c1, c2\ne="
        self.assertEqual(
            expected,
            utils.format_dict_of_list(
                {
                    'a': ['a2', 'a1'],
                    'b': ['b2', 'b1'],
                    'c': ['c1', 'c2'],
                    'd': None,
                    'e': [],
                },
                separator='\n',
            ),
        )
        self.assertEqual(
            expected,
            utils.format_dict_of_list(
                {
                    'c': ['c1', 'c2'],
                    'a': ['a2', 'a1'],
                    'b': ['b2', 'b1'],
                    'e': [],
                },
                separator='\n',
            ),
        )
        self.assertIsNone(utils.format_dict_of_list(None, separator='\n'))

    def test_format_list(self):
        expected = 'a, b, c'
        self.assertEqual(expected, utils.format_list(['a', 'b', 'c']))
        self.assertEqual(expected, utils.format_list(['c', 'b', 'a']))
        self.assertIsNone(utils.format_list(None))

    def test_format_list_of_dicts(self):
        expected = "a='b', c='d'\ne='f'"
        sorted_data = [{'a': 'b', 'c': 'd'}, {'e': 'f'}]
        unsorted_data = [{'c': 'd', 'a': 'b'}, {'e': 'f'}]
        self.assertEqual(expected, utils.format_list_of_dicts(sorted_data))
        self.assertEqual(expected, utils.format_list_of_dicts(unsorted_data))
        self.assertEqual('', utils.format_list_of_dicts([]))
        self.assertEqual('', utils.format_list_of_dicts([{}]))
        self.assertIsNone(utils.format_list_of_dicts(None))

    def test_format_list_separator(self):
        expected = 'a\nb\nc'
        actual_pre_sorted = utils.format_list(['a', 'b', 'c'], separator='\n')
        actual_unsorted = utils.format_list(['c', 'b', 'a'], separator='\n')
        self.assertEqual(expected, actual_pre_sorted)
        self.assertEqual(expected, actual_unsorted)


class TestAssertItemEqual(base.TestCommand):
    def test_assert_normal_item(self):
        expected = ['a', 'b', 'c']
        actual = ['a', 'b', 'c']
        self.assertItemEqual(expected, actual)

    def test_assert_item_with_formattable_columns(self):
        expected = [
            format_columns.DictColumn({'a': 1, 'b': 2}),
            format_columns.ListColumn(['x', 'y', 'z']),
        ]
        actual = [
            format_columns.DictColumn({'a': 1, 'b': 2}),
            format_columns.ListColumn(['x', 'y', 'z']),
        ]
        self.assertItemEqual(expected, actual)

    def test_assert_item_different_length(self):
        expected = ['a', 'b', 'c']
        actual = ['a', 'b']
        self.assertRaises(
            AssertionError, self.assertItemEqual, expected, actual
        )

    def test_assert_item_formattable_columns_vs_legacy_formatter(self):
        expected = [
            format_columns.DictColumn({'a': 1, 'b': 2}),
            format_columns.ListColumn(['x', 'y', 'z']),
        ]
        actual = [
            utils.format_dict({'a': 1, 'b': 2}),
            utils.format_list(['x', 'y', 'z']),
        ]
        self.assertRaises(
            AssertionError, self.assertItemEqual, expected, actual
        )

    def test_assert_item_different_formattable_columns(self):
        class ExceptionColumn(cliff_columns.FormattableColumn):
            def human_readable(self):
                raise Exception('always fail')

        expected = [format_columns.DictColumn({'a': 1, 'b': 2})]
        actual = [ExceptionColumn({'a': 1, 'b': 2})]
        # AssertionError is a subclass of Exception
        # so raising AssertionError ensures ExceptionColumn.human_readable()
        # is not called.
        self.assertRaises(
            AssertionError, self.assertItemEqual, expected, actual
        )

    def test_assert_list_item(self):
        expected = [
            ['a', 'b', 'c'],
            [
                format_columns.DictColumn({'a': 1, 'b': 2}),
                format_columns.ListColumn(['x', 'y', 'z']),
            ],
        ]
        actual = [
            ['a', 'b', 'c'],
            [
                format_columns.DictColumn({'a': 1, 'b': 2}),
                format_columns.ListColumn(['x', 'y', 'z']),
            ],
        ]
        self.assertListItemEqual(expected, actual)


class TestSDKUtils(base.TestCase):
    def setUp(self):
        super().setUp()

    def _test_get_osc_show_columns_for_sdk_resource(
        self,
        sdk_resource,
        column_map,
        expected_display_columns,
        expected_attr_columns,
    ):
        (
            display_columns,
            attr_columns,
        ) = utils.get_osc_show_columns_for_sdk_resource(
            sdk_resource, column_map
        )
        self.assertEqual(expected_display_columns, display_columns)
        self.assertEqual(expected_attr_columns, attr_columns)

    def test_get_osc_show_columns_for_sdk_resource_empty(self):
        self._test_get_osc_show_columns_for_sdk_resource(
            {}, {}, tuple(), tuple()
        )

    def test_get_osc_show_columns_for_sdk_resource_empty_map(self):
        self._test_get_osc_show_columns_for_sdk_resource(
            {'foo': 'foo1'}, {}, ('foo',), ('foo',)
        )

    def test_get_osc_show_columns_for_sdk_resource_empty_data(self):
        self._test_get_osc_show_columns_for_sdk_resource(
            {}, {'foo': 'foo_map'}, ('foo_map',), ('foo_map',)
        )

    def test_get_osc_show_columns_for_sdk_resource_map(self):
        self._test_get_osc_show_columns_for_sdk_resource(
            {'foo': 'foo1'}, {'foo': 'foo_map'}, ('foo_map',), ('foo',)
        )

    def test_get_osc_show_columns_for_sdk_resource_map_dup(self):
        self._test_get_osc_show_columns_for_sdk_resource(
            {'foo': 'foo1', 'foo_map': 'foo1'},
            {'foo': 'foo_map'},
            ('foo_map',),
            ('foo',),
        )

    def test_get_osc_show_columns_for_sdk_resource_map_full(self):
        self._test_get_osc_show_columns_for_sdk_resource(
            {'foo': 'foo1', 'bar': 'bar1'},
            {'foo': 'foo_map', 'new': 'bar'},
            ('bar', 'foo_map'),
            ('bar', 'foo'),
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Authentication Library"""

import argparse
from typing import TYPE_CHECKING, TypedDict, TypeVar

from keystoneauth1.identity import base as identity_base
from keystoneauth1.identity.v3 import k2k
from keystoneauth1.loading import base

from osc_lib import exceptions as exc
from osc_lib.i18n import _
from osc_lib import utils

if TYPE_CHECKING:
    from openstack.config import cloud_region


# Initialize the list of Authentication plugins early in order
# to get the command-line options
PLUGIN_LIST: frozenset[str] | None = None


class _OptionDict(TypedDict):
    env: str
    help: str


# List of plugin command line options
OPTIONS_LIST: dict[str, _OptionDict] = {}


def get_plugin_list() -> frozenset[str]:
    """Gather plugin list and cache it"""

    global PLUGIN_LIST

    if PLUGIN_LIST is None:
        PLUGIN_LIST = base.get_available_plugin_names()
    return PLUGIN_LIST


def get_options_list() -> dict[str, _OptionDict]:
    """Gather plugin options so the help action has them available"""

    global OPTIONS_LIST

    if not OPTIONS_LIST:
        for plugin_name in get_plugin_list():
            plugin_options = base.get_plugin_options(plugin_name)
            for o in plugin_options:
                os_name = o.name.lower().replace('_', '-')
                os_env_name = 'OS_' + os_name.upper().replace('-', '_')
                OPTIONS_LIST.setdefault(
                    os_name,
                    {'env': os_env_name, 'help': ''},
                )
                # TODO(mhu) simplistic approach, would be better to only add
                # help texts if they vary from one auth plugin to another
                # also the text rendering is ugly in the CLI ...
                OPTIONS_LIST[os_name]['help'] += (
                    f'With {plugin_name}: {o.help}\n'
                )
    return OPTIONS_LIST


def check_valid_authorization_options(
    options: 'cloud_region.CloudRegion',
    auth_plugin_name: str,
) -> None:
    """Validate authorization options, and provide helpful error messages."""
    if (
        options.auth.get('project_id')
        and not options.auth.get('domain_id')
        and not options.auth.get('domain_name')
        and not options.auth.get('project_name')
        and not options.auth.get('tenant_id')
        and not options.auth.get('tenant_name')
    ):
        raise exc.CommandError(
            _(
                'Missing parameter(s): '
                'Set either a project or a domain scope, but not both. Set a '
                'project scope with --os-project-name, OS_PROJECT_NAME, or '
                'auth.project_name. Alternatively, set a domain scope with '
                '--os-domain-name, OS_DOMAIN_NAME or auth.domain_name.'
            )
        )


def check_valid_authentication_options(
    options: 'cloud_region.CloudRegion',
    auth_plugin_name: str,
) -> None:
    """Validate authentication options, and provide helpful error messages."""
    # Get all the options defined within the plugin.
    plugin_opts = {
        opt.dest: opt for opt in base.get_plugin_options(auth_plugin_name)
    }

    # NOTE(aloga): this is an horrible hack. We need a way to specify the
    # required options in the plugins. Using the "required" argument for
    # the oslo_config.cfg.Opt does not work, as it is not possible to load the
    # plugin if the option is not defined, so the error will simply be:
    # "NoMatchingPlugin: The plugin foobar could not be found"
    msgs = []

    # when no auth params are passed in, user advised to use os-cloud
    if not options.auth and auth_plugin_name != 'none':
        msgs.append(_('Set a cloud-name with --os-cloud or OS_CLOUD'))
    else:
        if 'password' in plugin_opts and not (
            options.auth.get('username') or options.auth.get('user_id')
        ):
            msgs.append(
                _(
                    'Set a username with --os-username, OS_USERNAME,'
                    ' or auth.username'
                    ' or set a user-id with --os-user-id, OS_USER_ID,'
                    ' or auth.user_id'
                )
            )
        if 'auth_url' in plugin_opts and not options.auth.get('auth_url'):
            msgs.append(
                _(
                    'Set an authentication URL, with --os-auth-url,'
                    ' OS_AUTH_URL or auth.auth_url'
                )
            )
        if 'url' in plugin_opts and not options.auth.get('url'):
            msgs.append(
                _('Set a service URL, with --os-url, OS_URL or auth.url')
            )
        if 'token' in plugin_opts and not options.auth.get('token'):
            msgs.append(
                _('Set a token with --os-token, OS_TOKEN or auth.token')
            )

    if msgs:
        raise exc.CommandError(
            _('Missing parameter(s): \n%s') % '\n'.join(msgs)
        )


_ArgumentParserT = TypeVar('_ArgumentParserT', bound=argparse.ArgumentParser)


def build_auth_plugins_option_parser(
    parser: _ArgumentParserT,
) -> _ArgumentParserT:
    """Auth plugins options builder

    Builds dynamically the list of options expected by each available
    authentication plugin.
    """
    available_plugins = list(get_plugin_list())
    parser.add_argument(
        '--os-auth-type',
        metavar='<auth-type>',
        dest='auth_type',
        default=utils.env('OS_AUTH_TYPE'),
        help=_(
            'Select an authentication type. Available types: %s.'
            ' Default: selected based on --os-username/--os-token'
            ' (Env: OS_AUTH_TYPE)'
        )
        % ', '.join(available_plugins),
        choices=available_plugins,
    )
    # Maintain compatibility with old tenant env vars
    envs = {
        'OS_PROJECT_NAME': utils.env(
            'OS_PROJECT_NAME', default=utils.env('OS_TENANT_NAME')
        ),
        'OS_PROJECT_ID': utils.env(
            'OS_PROJECT_ID', default=utils.env('OS_TENANT_ID')
        ),
    }
    for o in get_options_list():
        # Remove tenant options from KSC plugins and replace them below
        if 'tenant' not in o:
            parser.add_argument(
                '--os-' + o,
                metavar=f'<auth-{o}>',
                dest=o.replace('-', '_'),
                default=envs.get(
                    OPTIONS_LIST[o]['env'],
                    utils.env(OPTIONS_LIST[o]['env']),
                ),
                help=_('%(help)s\n(Env: %(env)s)')
                % {
                    'help': OPTIONS_LIST[o]['help'],
                    'env': OPTIONS_LIST[o]['env'],
                },
            )
    # add tenant-related options for compatibility
    # this is deprecated but still used in some tempest tests...
    parser.add_argument(
        '--os-tenant-name',
        metavar='<auth-tenant-name>',
        dest='os_project_name',
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        '--os-tenant-id',
        metavar='<auth-tenant-id>',
        dest='os_project_id',
        help=argparse.SUPPRESS,
    )
    return parser


def get_keystone2keystone_auth(
    local_auth: identity_base.BaseIdentityPlugin,
    service_provider: str,
    project_id: str | None = None,
    project_name: str | None = None,
    project_domain_id: str | None = None,
    project_domain_name: str | None = None,
) -> k2k.Keystone2Keystone:
    """Return Keystone 2 Keystone authentication for service provider.

    :param local_auth: authentication to use with the local Keystone
    :param service_provider: service provider id as registered in Keystone
    :param project_id: project id to scope to in the service provider
    :param project_name: project name to scope to in the service provider
    :param project_domain_id: id of domain in the service provider
    :param project_domain_name: name of domain to in the service provider
    :return: Keystone2Keystone auth object for service provider
    """
    return k2k.Keystone2Keystone(
        local_auth,
        service_provider,
        project_id=project_id,
        project_name=project_name,
        project_domain_id=project_domain_id,
        project_domain_name=project_domain_name,
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import Iterable, Mapping
import json
import sys
from typing import Any, Protocol
from unittest import mock

import requests

AUTH_URL = "https://example.com/identity"
USERNAME = 'itchy'
USER_ID = '2354b7c1-f681-4c39-8003-4fe9d1eabb65'
PASSWORD = 'scratchy'  # noqa: S105
PROJECT_NAME = 'poochie'
PROJECT_ID = '30c3da29-61f5-4b7b-8eb2-3d18287428c7'
REGION_NAME = 'richie'
INTERFACE = 'catchy'
TOKEN = 'foobar'  # noqa: S105
IDENTITY_API_VERSION = '3'


class FakeStdout:
    def __init__(self) -> None:
        self.content: list[str] = []

    def write(self, text: str) -> None:
        self.content.append(text)

    def make_string(self) -> str:
        result = ''
        for line in self.content:
            result = result + line
        return result


# TODO(stephenfin): Replace with io.Writer when Python 3.14 is our minimum
# supported version
class IOWriterProtocol(Protocol):
    def write(self, text: str) -> None: ...


class LoggerProtocol(Protocol):
    def debug(self, msg: str, *args: Any, **kwargs: Any) -> None: ...

    def info(self, msg: str, *args: Any, **kwargs: Any) -> None: ...

    def warning(self, msg: str, *args: Any, **kwargs: Any) -> None: ...

    def error(self, msg: str, *args: Any, **kwargs: Any) -> None: ...

    def critical(self, msg: str, *args: Any, **kwargs: Any) -> None: ...


class FakeLog:
    def __init__(self) -> None:
        self.messages: dict[str, str] = {}

    def debug(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.messages['debug'] = msg

    def info(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.messages['info'] = msg

    def warning(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.messages['warning'] = msg

    def error(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.messages['error'] = msg

    def critical(self, msg: str, *args: Any, **kwargs: Any) -> None:
        self.messages['critical'] = msg


class FakeApp:
    client_manager: 'FakeClientManager'

    def __init__(self, _stdout: IOWriterProtocol, _log: LoggerProtocol):
        self.api_version: dict[str, str] = {}
        self.stdin = sys.stdin
        self.stdout = _stdout or sys.stdout
        self.stderr = sys.stderr
        self.log = _log


class FakeOptions:
    def __init__(self, **kwargs: Any) -> None:
        self.os_beta_command = False


class FakeClientManager:
    def __init__(self) -> None:
        self.compute = None
        self.identity = None
        self.image = None
        self.object_store = None
        self.volume = None
        self.network = None
        self.sdk_connection = mock.Mock()

        self.session = None
        self.auth_ref = None
        self.auth_plugin_name = None

    def get_configuration(self) -> dict[str, Any]:
        return {
            'auth': {
                'username': USERNAME,
                'password': PASSWORD,
                'token': TOKEN,
            },
            'region': REGION_NAME,
            'identity_api_version': IDENTITY_API_VERSION,
        }


class FakeModule:
    def __init__(self, name: str, version: str) -> None:
        self.name = name
        self.__version__ = version
        # Workaround for openstacksdk case
        self.version = mock.Mock()
        self.version.__version__ = version


class FakeResource:
    def __init__(
        self,
        manager: Any = None,
        info: dict[str, Any] | None = None,
        loaded: bool = False,
        methods: dict[str, Any] | None = None,
    ) -> None:
        """Set attributes and methods for a resource.

        :param manager: The resource manager
        :param info: A dictionary with all attributes
        :param loaded: True if the resource is loaded in memory
        :param methods: A dictionary with all methods
        """
        info = info or {}
        methods = methods or {}

        self.__name__ = type(self).__name__
        self.manager = manager
        self._info = info
        self._add_details(info)
        self._add_methods(methods)
        self._loaded = loaded

    def _add_details(self, info: dict[str, Any]) -> None:
        for k, v in info.items():
            setattr(self, k, v)

    def _add_methods(self, methods: dict[str, Any]) -> None:
        """Fake methods with MagicMock objects.

        For each <@key, @value> pairs in methods, add an callable MagicMock
        object named @key as an attribute, and set the mock's return_value to
        @value. When users access the attribute with (), @value will be
        returned, which looks like a function call.
        """
        for name, ret in methods.items():
            method = mock.MagicMock(return_value=ret)
            setattr(self, name, method)

    def __repr__(self) -> str:
        reprkeys = sorted(
            k for k in self.__dict__.keys() if k[0] != '_' and k != 'manager'
        )
        info = ", ".join(f"{k}={getattr(self, k)}" for k in reprkeys)
        return f"<{self.__class__.__name__} {info}>"

    def keys(self) -> Iterable[str]:
        return self._info.keys()

    def to_dict(self) -> dict[str, Any]:
        return self._info

    @property
    def info(self) -> dict[str, Any]:
        return self._info

    def __getitem__(self, item: str) -> Any:
        return self._info.get(item)

    def get(self, item: str, default: Any = None) -> Any:
        return self._info.get(item, default)

    def pop(self, key: str, default_value: Any = None) -> Any:
        return self.info.pop(key, default_value)


class FakeResponse(requests.Response):
    def __init__(
        self,
        headers: Mapping[str, str] | None = None,
        status_code: int = 200,
        data: object | None = None,
        encoding: type[str] | type[bytes] | None = None,
    ) -> None:
        super().__init__()

        headers = headers or {}

        self.status_code = status_code

        self.headers.update(headers)
        self._content = json.dumps(data).encode()



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import copy
import json as jsonutils
from unittest import mock

from keystoneauth1.access import service_catalog
from keystoneauth1 import exceptions as ksa_exceptions
from keystoneauth1 import fixture as ksa_fixture
from keystoneauth1.identity import generic as generic_plugin
from keystoneauth1.identity.v3 import k2k
from keystoneauth1 import loading
from keystoneauth1 import noauth
from keystoneauth1 import token_endpoint
from openstack.config import cloud_config
from openstack.config import cloud_region
from openstack.config import defaults
from openstack import connection
from requests_mock.contrib import fixture as requests_mock_fixture

from osc_lib.api import auth
from osc_lib import clientmanager
from osc_lib import exceptions as exc
from osc_lib.test import base
from osc_lib.test import fakes

SERVICE_PROVIDER_ID = "bob"

TEST_RESPONSE_DICT = ksa_fixture.V2Token(
    token_id=fakes.TOKEN, user_name=fakes.USERNAME
)
_s = TEST_RESPONSE_DICT.add_service('identity', name='keystone')
_s.add_endpoint(fakes.AUTH_URL + ':5000/v2.0')
_s = TEST_RESPONSE_DICT.add_service('network', name='neutron')
_s.add_endpoint(fakes.AUTH_URL + ':9696')
_s = TEST_RESPONSE_DICT.add_service('compute', name='nova')
_s.add_endpoint(fakes.AUTH_URL + ':8774/v2')
_s = TEST_RESPONSE_DICT.add_service('image', name='glance')
_s.add_endpoint(fakes.AUTH_URL + ':9292')
_s = TEST_RESPONSE_DICT.add_service('object', name='swift')
_s.add_endpoint(fakes.AUTH_URL + ':8080/v1')

TEST_RESPONSE_DICT_V3 = ksa_fixture.V3Token(user_name=fakes.USERNAME)
TEST_RESPONSE_DICT_V3.set_project_scope()

TEST_VERSIONS = ksa_fixture.DiscoveryList(href=fakes.AUTH_URL)

AUTH_REF = {'version': 'v2.0'}
AUTH_REF.update(TEST_RESPONSE_DICT['access'])
SERVICE_CATALOG = service_catalog.ServiceCatalogV2(AUTH_REF)

AUTH_DICT = {
    'auth_url': fakes.AUTH_URL,
    'username': fakes.USERNAME,
    'password': fakes.PASSWORD,
    'project_name': fakes.PROJECT_NAME,
}


# This is deferred in api.auth but we need it here...
auth.get_options_list()


class Container:
    attr = clientmanager.ClientCache(lambda x: object())
    buggy_attr = clientmanager.ClientCache(lambda x: x.foo)

    def __init__(self):
        pass


class TestClientCache(base.TestCase):
    def test_singleton(self):
        # NOTE(dtroyer): Verify that the ClientCache descriptor only invokes
        # the factory one time and always returns the same value after that.
        c = Container()
        self.assertEqual(c.attr, c.attr)

    def test_attribute_error_propagates(self):
        c = Container()
        err = self.assertRaises(
            exc.PluginAttributeError, getattr, c, 'buggy_attr'
        )
        self.assertNotIsInstance(err, AttributeError)
        self.assertEqual("'Container' object has no attribute 'foo'", str(err))


class BaseTestClientManager(base.TestCase):
    """ClientManager class test framework"""

    default_password_auth = {
        'auth_url': fakes.AUTH_URL,
        'username': fakes.USERNAME,
        'password': fakes.PASSWORD,
        'project_name': fakes.PROJECT_NAME,
    }
    default_token_auth = {
        'auth_url': fakes.AUTH_URL,
        'token': fakes.TOKEN,
    }

    def setUp(self):
        super().setUp()
        self.mock = mock.Mock()
        self.requests = self.useFixture(requests_mock_fixture.Fixture())
        # fake v2password token retrieval
        self.stub_auth(json=TEST_RESPONSE_DICT)
        # fake token and token_endpoint retrieval
        self.stub_auth(
            json=TEST_RESPONSE_DICT,
            url='/'.join([fakes.AUTH_URL, 'v2.0/tokens']),
        )
        # fake v3password token retrieval
        self.stub_auth(
            json=TEST_RESPONSE_DICT_V3,
            url='/'.join([fakes.AUTH_URL, 'v3/auth/tokens']),
        )
        # fake password token retrieval
        self.stub_auth(
            json=TEST_RESPONSE_DICT_V3,
            url='/'.join([fakes.AUTH_URL, 'auth/tokens']),
        )
        # fake password version endpoint discovery
        self.stub_auth(json=TEST_VERSIONS, url=fakes.AUTH_URL, verb='GET')

        # Mock the auth plugin
        self.auth_mock = mock.Mock()

    def stub_auth(self, json=None, url=None, verb=None, **kwargs):
        subject_token = fakes.TOKEN
        base_url = fakes.AUTH_URL
        if json:
            text = jsonutils.dumps(json)
            headers = {
                'X-Subject-Token': subject_token,
                'Content-Type': 'application/json',
            }
        if not url:
            url = '/'.join([base_url, 'tokens'])
        url = url.replace("/?", "?")
        if not verb:
            verb = 'POST'
        self.requests.register_uri(
            verb,
            url,
            headers=headers,
            text=text,
        )

    def _clientmanager_class(self):
        """Allow subclasses to override the ClientManager class"""
        return clientmanager.ClientManager

    def _make_clientmanager(
        self,
        auth_args=None,
        config_args=None,
        identity_api_version=None,
        auth_plugin_name=None,
        auth_required=None,
    ):
        if identity_api_version is None:
            identity_api_version = '2.0'
        if auth_plugin_name is None:
            auth_plugin_name = 'password'

        if auth_plugin_name.endswith('password'):
            auth_dict = copy.deepcopy(self.default_password_auth)
        elif auth_plugin_name.endswith('token'):
            auth_dict = copy.deepcopy(self.default_token_auth)
        else:
            auth_dict = {}

        if auth_args is not None:
            auth_dict = auth_args

        cli_options = defaults.get_defaults()
        cli_options.update(
            {
                'auth_type': auth_plugin_name,
                'auth': auth_dict,
                'interface': fakes.INTERFACE,
                'region_name': fakes.REGION_NAME,
            }
        )
        if config_args is not None:
            cli_options.update(config_args)

        loader = loading.get_plugin_loader(auth_plugin_name)
        auth_plugin = loader.load_from_options(**auth_dict)
        client_manager = self._clientmanager_class()(
            cli_options=cloud_region.CloudRegion(
                name='t1',
                region_name='1',
                config=cli_options,
                auth_plugin=auth_plugin,
            ),
            api_version={
                'identity': identity_api_version,
            },
        )
        client_manager._auth_required = auth_required is True
        client_manager.setup_auth()
        client_manager.auth_ref

        self.assertEqual(
            auth_plugin_name,
            client_manager.auth_plugin_name,
        )
        return client_manager


class TestClientManager(BaseTestClientManager):
    def test_client_manager_none(self):
        none_auth = {
            'endpoint': fakes.AUTH_URL,
        }
        client_manager = self._make_clientmanager(
            auth_args=none_auth,
            auth_plugin_name='none',
        )

        self.assertEqual(
            fakes.AUTH_URL,
            client_manager._cli_options.config['auth']['endpoint'],
        )
        self.assertIsInstance(
            client_manager.auth,
            noauth.NoAuth,
        )
        # Check that the endpoint option works as the override
        self.assertEqual(
            fakes.AUTH_URL,
            client_manager.get_endpoint_for_service_type('baremetal'),
        )

    def test_client_manager_admin_token(self):
        token_auth = {
            'endpoint': fakes.AUTH_URL,
            'token': fakes.TOKEN,
        }
        client_manager = self._make_clientmanager(
            auth_args=token_auth,
            auth_plugin_name='admin_token',
        )

        self.assertEqual(
            fakes.AUTH_URL,
            client_manager._cli_options.config['auth']['endpoint'],
        )
        self.assertEqual(
            fakes.TOKEN,
            client_manager.auth.get_token(None),
        )
        self.assertIsInstance(
            client_manager.auth,
            token_endpoint.Token,
        )
        # NOTE(dtroyer): This is intentionally not assertFalse() as the return
        #                value from is_service_available() may be == None
        self.assertNotEqual(
            False,
            client_manager.is_service_available('network'),
        )

    def test_client_manager_password(self):
        client_manager = self._make_clientmanager(
            auth_required=True,
        )

        self.assertEqual(
            fakes.AUTH_URL,
            client_manager._cli_options.config['auth']['auth_url'],
        )
        self.assertEqual(
            fakes.USERNAME,
            client_manager._cli_options.config['auth']['username'],
        )
        self.assertEqual(
            fakes.PASSWORD,
            client_manager._cli_options.config['auth']['password'],
        )
        self.assertIsInstance(
            client_manager.auth,
            generic_plugin.Password,
        )
        self.assertTrue(client_manager.verify)
        self.assertIsNone(client_manager.cert)

        # These need to stick around until the old-style clients are gone
        self.assertEqual(
            AUTH_REF.pop('version'),
            client_manager.auth_ref.version,
        )
        self.assertEqual(
            AUTH_REF,
            client_manager.auth_ref._data['access'],
        )
        self.assertEqual(
            dir(SERVICE_CATALOG),
            dir(client_manager.auth_ref.service_catalog),
        )
        self.assertTrue(client_manager.is_service_available('network'))

    def test_client_manager_password_verify(self):
        client_manager = self._make_clientmanager(
            auth_required=True,
        )

        self.assertTrue(client_manager.verify)
        self.assertIsNone(client_manager.cacert)
        self.assertTrue(client_manager.is_service_available('network'))

    def test_client_manager_password_verify_ca(self):
        config_args = {
            'cacert': 'cafile',
        }
        client_manager = self._make_clientmanager(
            config_args=config_args,
            auth_required=True,
        )

        # Test that client_manager.verify is Requests-compatible,
        # i.e. it contains the value of cafile here
        self.assertTrue(client_manager.verify)
        self.assertEqual('cafile', client_manager.verify)
        self.assertEqual('cafile', client_manager.cacert)
        self.assertTrue(client_manager.is_service_available('network'))

    def test_client_manager_password_verify_false(self):
        config_args = {
            'verify': False,
        }
        client_manager = self._make_clientmanager(
            config_args=config_args,
            auth_required=True,
        )

        self.assertFalse(client_manager.verify)
        self.assertIsNone(client_manager.cacert)
        self.assertTrue(client_manager.is_service_available('network'))

    def test_client_manager_password_verify_insecure(self):
        config_args = {
            'insecure': True,
        }
        client_manager = self._make_clientmanager(
            config_args=config_args,
            auth_required=True,
        )

        self.assertFalse(client_manager.verify)
        self.assertIsNone(client_manager.cacert)
        self.assertTrue(client_manager.is_service_available('network'))

    def test_client_manager_password_verify_insecure_ca(self):
        config_args = {
            'insecure': True,
            'cacert': 'cafile',
        }
        client_manager = self._make_clientmanager(
            config_args=config_args,
            auth_required=True,
        )

        # insecure overrides cacert
        self.assertFalse(client_manager.verify)
        self.assertIsNone(client_manager.cacert)
        self.assertTrue(client_manager.is_service_available('network'))

    def test_client_manager_password_client_cert(self):
        config_args = {
            'cert': 'cert',
        }
        client_manager = self._make_clientmanager(
            config_args=config_args,
        )

        self.assertEqual('cert', client_manager.cert)

    def test_client_manager_password_client_key(self):
        config_args = {
            'cert': 'cert',
            'key': 'key',
        }
        client_manager = self._make_clientmanager(
            config_args=config_args,
        )

        self.assertEqual(('cert', 'key'), client_manager.cert)

    def test_client_manager_select_auth_plugin_password(self):
        # test password auth
        auth_args = {
            'auth_url': fakes.AUTH_URL,
            'username': fakes.USERNAME,
            'password': fakes.PASSWORD,
            'tenant_name': fakes.PROJECT_NAME,
        }
        self._make_clientmanager(
            auth_args=auth_args,
            identity_api_version='2.0',
            auth_plugin_name='v2password',
        )

        auth_args = copy.deepcopy(self.default_password_auth)
        auth_args.update(
            {
                'user_domain_name': 'default',
                'project_domain_name': 'default',
            }
        )
        self._make_clientmanager(
            auth_args=auth_args,
            identity_api_version='3',
            auth_plugin_name='v3password',
        )

        # Use v2.0 auth args
        auth_args = {
            'auth_url': fakes.AUTH_URL,
            'username': fakes.USERNAME,
            'password': fakes.PASSWORD,
            'tenant_name': fakes.PROJECT_NAME,
        }
        self._make_clientmanager(
            auth_args=auth_args,
            identity_api_version='2.0',
        )

        # Use v3 auth args
        auth_args = copy.deepcopy(self.default_password_auth)
        auth_args.update(
            {
                'user_domain_name': 'default',
                'project_domain_name': 'default',
            }
        )
        self._make_clientmanager(
            auth_args=auth_args,
            identity_api_version='3',
        )

        auth_args = copy.deepcopy(self.default_password_auth)
        auth_args.pop('username')
        auth_args.update({'user_id': fakes.USER_ID})
        self._make_clientmanager(
            auth_args=auth_args,
            identity_api_version='3',
        )

    def test_client_manager_select_auth_plugin_token(self):
        # test token auth
        self._make_clientmanager(
            # auth_args=auth_args,
            identity_api_version='2.0',
            auth_plugin_name='v2token',
        )
        self._make_clientmanager(
            # auth_args=auth_args,
            identity_api_version='3',
            auth_plugin_name='v3token',
        )
        self._make_clientmanager(
            # auth_args=auth_args,
            identity_api_version='x',
            auth_plugin_name='token',
        )

    def test_client_manager_select_auth_plugin_failure(self):
        self.assertRaises(
            ksa_exceptions.NoMatchingPlugin,
            self._make_clientmanager,
            identity_api_version='3',
            auth_plugin_name='bad_plugin',
        )

    @mock.patch('osc_lib.api.auth.check_valid_authentication_options')
    def test_client_manager_auth_setup_once(self, check_authn_options_func):
        loader = loading.get_plugin_loader('password')
        auth_plugin = loader.load_from_options(**AUTH_DICT)
        cli_options = defaults.get_defaults()
        cli_options.update(
            {
                'auth_type': 'password',
                'auth': AUTH_DICT,
                'interface': fakes.INTERFACE,
                'region_name': fakes.REGION_NAME,
            }
        )
        client_manager = self._clientmanager_class()(
            cli_options=cloud_config.CloudConfig(
                name='t1',
                region='1',
                config=cli_options,
                auth_plugin=auth_plugin,
            ),
            api_version={
                'identity': '2.0',
            },
        )
        self.assertFalse(client_manager._auth_setup_completed)
        client_manager.setup_auth()
        self.assertTrue(check_authn_options_func.called)
        self.assertTrue(client_manager._auth_setup_completed)

        # now make sure we don't do auth setup the second time around
        # by checking whether check_valid_auth_options() gets called again
        check_authn_options_func.reset_mock()
        client_manager.auth_ref
        check_authn_options_func.assert_not_called()

    def test_client_manager_endpoint_disabled(self):
        auth_args = copy.deepcopy(self.default_password_auth)
        auth_args.update(
            {
                'user_domain_name': 'default',
                'project_domain_name': 'default',
            }
        )
        # v3 fake doesn't have network endpoint
        client_manager = self._make_clientmanager(
            auth_args=auth_args,
            identity_api_version='3',
            auth_plugin_name='v3password',
        )

        self.assertFalse(client_manager.is_service_available('network'))

    def test_client_manager_k2k_auth_setup(self):
        loader = loading.get_plugin_loader('password')
        auth_plugin = loader.load_from_options(**AUTH_DICT)
        cli_options = defaults.get_defaults()
        cli_options.update(
            {
                'auth_type': 'password',
                'auth': AUTH_DICT,
                'interface': fakes.INTERFACE,
                'region_name': fakes.REGION_NAME,
                'service_provider': SERVICE_PROVIDER_ID,
                'remote_project_id': fakes.PROJECT_ID,
            }
        )
        client_manager = self._clientmanager_class()(
            cli_options=cloud_config.CloudConfig(
                name='t1',
                region='1',
                config=cli_options,
                auth_plugin=auth_plugin,
            ),
            api_version={
                'identity': '3',
            },
        )

        self.assertFalse(client_manager._auth_setup_completed)
        client_manager.setup_auth()
        # Note(knikolla): Make sure that the auth object is of the correct
        # type and that the service_provider is correctly set.
        self.assertIsInstance(client_manager.auth, k2k.Keystone2Keystone)
        self.assertEqual(client_manager.auth._sp_id, SERVICE_PROVIDER_ID)
        self.assertEqual(client_manager.auth.project_id, fakes.PROJECT_ID)
        self.assertTrue(client_manager._auth_setup_completed)

    def test_client_manager_none_auth(self):
        # test token auth
        client_manager = self._make_clientmanager(
            auth_args={},
            auth_plugin_name='none',
        )
        self.assertIsNone(
            client_manager.get_endpoint_for_service_type('compute')
        )

    def test_client_manager_endpoint_override(self):
        # test token auth
        client_manager = self._make_clientmanager(
            auth_args={},
            config_args={
                'compute_endpoint_override': 'http://example.com',
                'foo_bar_endpoint_override': 'http://example2.com',
            },
            auth_plugin_name='none',
        )
        self.assertEqual(
            'http://example.com',
            client_manager.get_endpoint_for_service_type('compute'),
        )
        self.assertEqual(
            'http://example2.com',
            client_manager.get_endpoint_for_service_type('foo-bar'),
        )
        self.assertTrue(client_manager.is_service_available('compute'))

    def test_client_manager_connection(self):
        client_manager = self._make_clientmanager(
            auth_required=True,
        )

        self.assertIsInstance(
            client_manager.sdk_connection,
            connection.Connection,
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import logging
from unittest import mock

from osc_lib import logs
from osc_lib.test import base


class TestContext(base.TestCase):
    def test_log_level_from_options(self):
        opts = mock.Mock()
        opts.verbose_level = 0
        self.assertEqual(logging.ERROR, logs.log_level_from_options(opts))
        opts.verbose_level = 1
        self.assertEqual(logging.WARNING, logs.log_level_from_options(opts))
        opts.verbose_level = 2
        self.assertEqual(logging.INFO, logs.log_level_from_options(opts))
        opts.verbose_level = 3
        self.assertEqual(logging.DEBUG, logs.log_level_from_options(opts))

    def test_log_level_from_config(self):
        cfg = {'verbose_level': 0}
        self.assertEqual(logging.ERROR, logs.log_level_from_config(cfg))
        cfg = {'verbose_level': 1}
        self.assertEqual(logging.WARNING, logs.log_level_from_config(cfg))
        cfg = {'verbose_level': 2}
        self.assertEqual(logging.INFO, logs.log_level_from_config(cfg))
        cfg = {'verbose_level': 3}
        self.assertEqual(logging.DEBUG, logs.log_level_from_config(cfg))
        cfg = {'verbose_level': 1, 'log_level': 'critical'}
        self.assertEqual(logging.CRITICAL, logs.log_level_from_config(cfg))
        cfg = {'verbose_level': 1, 'log_level': 'error'}
        self.assertEqual(logging.ERROR, logs.log_level_from_config(cfg))
        cfg = {'verbose_level': 1, 'log_level': 'warning'}
        self.assertEqual(logging.WARNING, logs.log_level_from_config(cfg))
        cfg = {'verbose_level': 1, 'log_level': 'info'}
        self.assertEqual(logging.INFO, logs.log_level_from_config(cfg))
        cfg = {'verbose_level': 1, 'log_level': 'debug'}
        self.assertEqual(logging.DEBUG, logs.log_level_from_config(cfg))
        cfg = {'verbose_level': 1, 'log_level': 'bogus'}
        self.assertEqual(logging.WARNING, logs.log_level_from_config(cfg))
        cfg = {'verbose_level': 1, 'log_level': 'info', 'debug': True}
        self.assertEqual(logging.DEBUG, logs.log_level_from_config(cfg))

    @mock.patch('warnings.simplefilter')
    def test_set_warning_filter(self, simplefilter):
        logs.set_warning_filter(logging.ERROR)
        simplefilter.assert_called_with("ignore")
        logs.set_warning_filter(logging.WARNING)
        simplefilter.assert_called_with("ignore")
        logs.set_warning_filter(logging.INFO)
        simplefilter.assert_called_with("once")


class TestFileFormatter(base.TestCase):
    def test_nothing(self):
        formatter = logs._FileFormatter()
        self.assertEqual(
            (
                '%(asctime)s.%(msecs)03d %(process)d %(levelname)s '
                '%(name)s %(message)s'
            ),
            formatter.fmt,
        )

    def test_options(self):
        class Opts:
            cloud = 'cloudy'
            os_project_name = 'projecty'
            username = 'usernamey'

        options = Opts()
        formatter = logs._FileFormatter(options=options)
        self.assertEqual(
            (
                '%(asctime)s.%(msecs)03d %(process)d %(levelname)s '
                '%(name)s [cloudy usernamey projecty] %(message)s'
            ),
            formatter.fmt,
        )

    def test_config(self):
        config = mock.Mock()
        config.config = {'cloud': 'cloudy'}
        config.auth = {'project_name': 'projecty', 'username': 'usernamey'}
        formatter = logs._FileFormatter(config=config)
        self.assertEqual(
            (
                '%(asctime)s.%(msecs)03d %(process)d %(levelname)s '
                '%(name)s [cloudy usernamey projecty] %(message)s'
            ),
            formatter.fmt,
        )


class TestLogConfigurator(base.TestCase):
    def setUp(self):
        super().setUp()
        self.options = mock.Mock()
        self.options.verbose_level = 1
        self.options.log_file = None
        self.options.debug = False
        self.root_logger = mock.Mock()
        self.root_logger.setLevel = mock.Mock()
        self.root_logger.addHandler = mock.Mock()
        self.requests_log = mock.Mock()
        self.requests_log.setLevel = mock.Mock()
        self.cliff_log = mock.Mock()
        self.cliff_log.setLevel = mock.Mock()
        self.stevedore_log = mock.Mock()
        self.stevedore_log.setLevel = mock.Mock()
        self.iso8601_log = mock.Mock()
        self.iso8601_log.setLevel = mock.Mock()
        self.loggers = [
            self.root_logger,
            self.requests_log,
            self.cliff_log,
            self.stevedore_log,
            self.iso8601_log,
        ]

    @mock.patch('logging.StreamHandler')
    @mock.patch('logging.getLogger')
    @mock.patch('osc_lib.logs.set_warning_filter')
    def test_init(self, warning_filter, getLogger, handle):
        getLogger.side_effect = self.loggers
        console_logger = mock.Mock()
        console_logger.setFormatter = mock.Mock()
        console_logger.setLevel = mock.Mock()
        handle.return_value = console_logger

        configurator = logs.LogConfigurator(self.options)

        getLogger.assert_called_with('iso8601')  # last call
        warning_filter.assert_called_with(logging.WARNING)
        self.root_logger.setLevel.assert_called_with(logging.DEBUG)
        self.root_logger.addHandler.assert_called_with(console_logger)
        self.requests_log.setLevel.assert_called_with(logging.ERROR)
        self.cliff_log.setLevel.assert_called_with(logging.ERROR)
        self.stevedore_log.setLevel.assert_called_with(logging.ERROR)
        self.iso8601_log.setLevel.assert_called_with(logging.ERROR)
        self.assertFalse(configurator.dump_trace)

    @mock.patch('logging.getLogger')
    @mock.patch('osc_lib.logs.set_warning_filter')
    def test_init_no_debug(self, warning_filter, getLogger):
        getLogger.side_effect = self.loggers
        self.options.debug = True

        configurator = logs.LogConfigurator(self.options)

        warning_filter.assert_called_with(logging.DEBUG)
        self.requests_log.setLevel.assert_called_with(logging.DEBUG)
        self.assertTrue(configurator.dump_trace)

    @mock.patch('logging.FileHandler')
    @mock.patch('logging.getLogger')
    @mock.patch('osc_lib.logs.set_warning_filter')
    @mock.patch('osc_lib.logs._FileFormatter')
    def test_init_log_file(self, formatter, warning_filter, getLogger, handle):
        getLogger.side_effect = self.loggers
        self.options.log_file = '/tmp/log_file'
        file_logger = mock.Mock()
        file_logger.setFormatter = mock.Mock()
        file_logger.setLevel = mock.Mock()
        handle.return_value = file_logger
        mock_formatter = mock.Mock()
        formatter.return_value = mock_formatter

        logs.LogConfigurator(self.options)

        handle.assert_called_with(filename=self.options.log_file)
        self.root_logger.addHandler.assert_called_with(file_logger)
        file_logger.setFormatter.assert_called_with(mock_formatter)
        file_logger.setLevel.assert_called_with(logging.WARNING)

    @mock.patch('logging.FileHandler')
    @mock.patch('logging.getLogger')
    @mock.patch('osc_lib.logs.set_warning_filter')
    @mock.patch('osc_lib.logs._FileFormatter')
    def test_configure(self, formatter, warning_filter, getLogger, handle):
        getLogger.side_effect = self.loggers
        configurator = logs.LogConfigurator(self.options)
        cloud_config = mock.Mock()
        config_log = '/tmp/config_log'
        cloud_config.config = {
            'log_file': config_log,
            'verbose_level': 1,
            'log_level': 'info',
        }
        file_logger = mock.Mock()
        file_logger.setFormatter = mock.Mock()
        file_logger.setLevel = mock.Mock()
        handle.return_value = file_logger
        mock_formatter = mock.Mock()
        formatter.return_value = mock_formatter

        configurator.configure(cloud_config)

        warning_filter.assert_called_with(logging.INFO)
        handle.assert_called_with(filename=config_log)
        self.root_logger.addHandler.assert_called_with(file_logger)
        file_logger.setFormatter.assert_called_with(mock_formatter)
        file_logger.setLevel.assert_called_with(logging.INFO)
        self.assertFalse(configurator.dump_trace)
